# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
"""
Module for reading configuration files

Copyright (C) 2011 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import collections
import six
import os
import uuid

from six.moves.configparser import SafeConfigParser, NoOptionError, Error, MissingSectionHeaderError, ParsingError,\
DEFAULTSECT
from virtwho import log, SAT5, SAT6
from .password import Password, InvalidKeyFile
from binascii import unhexlify
from binascii import Error as BinasciiError
from . import util

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 doesn't have OrderedDict, we need to have our own
    from .util import OrderedDict

# Module-level logger
logger = log.getLogger(name='config', queue=False)

_effective_config = None

VW_CONF_DIR = "/etc/virt-who.d/"
VW_TYPES = ("libvirt", "vdsm", "esx", "rhevm", "hyperv", "fake", "xen", "kubevirt", "ahv")
VW_GENERAL_CONF_PATH = "/etc/virt-who.conf"
VW_GLOBAL = "global"
VW_VIRT_DEFAULTS_SECTION_NAME = "defaults"
VW_ENV_CLI_SECTION_NAME = "env/cmdline"

FILTER_TYPES = ("regex", "wildcards")

# Default interval for sending list of UUIDs
DefaultInterval = 3600  # One per hour
MinimumSendInterval = 60  # One minute


class InvalidOption(Error):
    pass


class InvalidPasswordFormat(Exception):
    pass


def parse_list(s):
    """
    Parse comma-separated list of items (that might be in quotes) to the list of strings
    """
    items = []

    read_to = None  # everything until next `read_to` (single or double
    # quotes) character is one item
    item = []  # characters of current items so far
    i = 0
    while i < len(s):
        ch = s[i]
        if ch == '\\':
            item.append(s[i + 1])
            i += 1
        elif read_to is None:
            if ch in ['"', "'"]:
                # everything until next quote is single item
                read_to = ch
            elif ch == ',':
                items.append(''.join(item))
                item = []
            elif not ch.isspace():
                item.append(ch)
        elif ch == read_to:
            read_to = None
        else:
            item.append(ch)
        i += 1
    if read_to is not None:
        raise ValueError("Unterminated %s sign" % {"'": "single quote", '"': "double quote"}.get(read_to, read_to))
    if item:
        items.append(''.join(item))
    return items


class NotSetSentinel(object):
    """
    An empty object subclass that is meant to be used in place of 'None'.
    We might want to set a config value to 'None'
    """
    pass


class Info(object):
    """
    A class containing keys with values.
    If the kwargs passed to the init of this class do not include all
    required_args, init will fail with a ValueError.

    All required and optional kwargs will be accessible on an instance of
    this object as attributes.
    No required or optional kwargs can start with an '_'.

    This class is used as a base for creating entities that are constituted
    entirely by a set of named values. In the case of virt-who we subclass
    this for all destination objects in order to attempt to pull all required
    info for a given type of destination out of configs and to help determine
    what unique destinations exist across all configurations.
    """
    required_kwargs = ()
    optional_kwargs = ()

    def __init__(self, **kwargs):
        self.__dict__["_options"] = {}
        attributes_to_add = []
        attributes_to_add.extend(type(self).required_kwargs)
        if type(self).optional_kwargs:
            attributes_to_add.extend(type(self).optional_kwargs)

        for arg in attributes_to_add:
            try:
                value = kwargs[arg]
                if value in [None, NotSetSentinel, '']:
                    raise TypeError
                self._options[arg] = value
            except (KeyError, TypeError):
                if arg in type(self).required_kwargs:
                    raise ValueError("Missing required option: %s" % arg)

    def __hash__(self):
        to_hash = []
        for item in self._options.values():
            if isinstance(item, list):
                to_hash.append(tuple(item))
            else:
                to_hash.append(item)
        return hash(tuple(to_hash))

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super(object, self).__setattr__(name, value)
            return
        self._options[name] = value

    def __getattr__(self, name):
        if name.startswith('_'):
            return super(object, self).__getattr__(name)
        item = self._options.get(name, NotSetSentinel)
        if item is NotSetSentinel:
            raise AttributeError("No attribute '%s'" % name)
        else:
            return item

    def __getitem__(self, item):
        try:
            return self.__dict__["_options"][item]
        except KeyError:
            pass
        return NotSetSentinel

    def __setitem__(self, key, value):
        self.__dict__["_options"][key] = value

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.__dict__['_options'] == self.__dict__['_options']

    def __ne__(self, other):
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented
        return not result

    def __iter__(self):
        for key, value in self.__dict__["_options"].items():
            yield (key, value)

    def keys(self):
        return list(self.__dict__['_options'].keys())


# Should this be defined in the manager that actually requires these values?
class Satellite5DestinationInfo(Info):
    required_kwargs = (
        "sat_server",
        "sat_username",
        "sat_password",
    )
    optional_kwargs = (
        "filter_hosts",
        "exclude_hosts",
    )


# Should this be defined in the manager that actually requires these values?
class Satellite6DestinationInfo(Info):
    required_kwargs = (
        "owner",
    )
    optional_kwargs = (
        "rhsm_hostname",
        "rhsm_port",
        "rhsm_prefix",
        "rhsm_username",
        "rhsm_password",
        "rhsm_proxy_hostname",
        "rhsm_proxy_port",
        "rhsm_proxy_user",
        "rhsm_proxy_password",
        "rhsm_no_proxy",
        "rhsm_insecure",
    )


class DefaultDestinationInfo(Info):
    pass


default_destination_info = DefaultDestinationInfo()
default_destination_info.name = "default_destination"


class StripQuotesConfigParser(SafeConfigParser):
    def get(self, section, option, **kwargs):
        # Don't call super, SafeConfigParser is not inherited from object
        value = SafeConfigParser.get(self, section, option, **kwargs)
        for quote in ('"', "'"):
            # Strip the quotes only when the value starts with quote,
            # ends with quote but doesn't contain it inside
            if value.startswith(quote) and value.endswith(quote) and quote not in value[1:-1]:
                return value.strip(quote)
        return value

    def _read(self, fp, fpname):
        """ If we're running python2, before reading the config file, parse through it to check if a commented out
        continuation line was detected (line starts with spaces/tabs, followed by '#') and if so, warn the user.
        Note: In python3, the configparser lib itself will remove commented out line continuations, so there is
        no need to perform this check.
        """
        if six.PY2:
            line_number = 0
            for line in fp:
                line_number = line_number + 1
                if line[0] == ' ' or line[0] == '\t':
                    if len(line.strip()) > 0 and line.strip()[0] == '#':
                        # warn the user that a continuation of the previous line is commented out
                        logger.warn("A line continuation (line starts with space) that is commented out "
                                    "was detected in file %s, line number %s.", fpname, line_number)
            fp.seek(0)
        SafeConfigParser._read(self, fp, fpname)


class DestinationToSourceMapper(object):
    def __init__(self, effective_config):
        self._configs = effective_config.virt_sections()
        self.effective_config = effective_config
        self.logger = logger
        self.sources = set()
        self.dests = set()
        self.dest_to_sources_map = {}
        self.update_dest_to_source_map()

    def update_dest_to_source_map(self):
        sources, dests, d_to_s, orphan_sources = \
            DestinationToSourceMapper.map_destinations_to_sources(self._configs)
        if orphan_sources:
            # Only use the default destination for sources that have a matching sm_type
            sources_to_default = [
                    orphan for orphan in orphan_sources \
                    if self.effective_config[orphan]['sm_type'] != SAT5
            ]
            if sources_to_default:
                d_to_s[default_destination_info] = sources_to_default
                dests.add(default_destination_info)
        self.sources = sources
        self.dests = dests
        self.dest_to_sources_map = d_to_s


    @staticmethod
    def map_destinations_to_sources(configs, dest_classes=(Satellite5DestinationInfo, Satellite6DestinationInfo)):
        """
        Create a mapping of destinations to sources, given all the collected
        and parsed configuration objects.

        @param configs: A list of Config objects to analyze
        @type configs: list

        @param dest_classes: An iterable of classes to try to pull out of a
        config object
        @type dest_classes: Iterable

        @rtype: (set, set, dict, set)
        """

        # Will include the names of the configs
        sources = set()
        sources_without_destinations = set()
        # Will include all dest info objects from all configs
        dests = set()
        dest_classes = dest_classes or (Satellite5DestinationInfo,
                                        Satellite6DestinationInfo)
        # We could expand this to include other specific sources
        # For now all configs are considered to define a source at a minimum
        # As such we will consider each config section name to be unique
        dest_to_source_map = {}  # The resultant
        # mapping
        # of destination to
        # sources
        for name, config in configs:
            sources.add(name)
            sources_without_destinations.add(name)
            for dest in DestinationToSourceMapper.parse_dests_from_dict(config,
                                                                        dest_classes):
                dests.add(dest)
                current_sources = dest_to_source_map.get(dest, set())
                current_sources.update(set([name]))
                sources_without_destinations.difference_update(set([name]))
                dest_to_source_map[dest] = current_sources
        for dest, source_set in dest_to_source_map.items():
            dest_to_source_map[dest] = sorted(list(source_set))
        sources_without_destinations = sorted(list(sources_without_destinations))
        return sources, dests, dest_to_source_map, sources_without_destinations

    @staticmethod
    def parse_dests_from_dict(dict_to_parse,
                              dest_classes=(
                                Satellite5DestinationInfo,
                                Satellite6DestinationInfo,
                              )):
        """
        @param dict_to_parse: The dict of kwargs to try to create a
        destination out of
        @type dict_to_parse: dict

        @param dest_classes: An iterable of Info classes to try to create
            based on the given dict of options
        @type dest_classes: collections.Iterable

        @return: A set of the info objects that could be created from the
        given dict
        @rtype: set
        """
        dests = set()
        for dest_class in dest_classes:
            dest = None
            try:
                dest = dest_class(**dict_to_parse)
            except ValueError:
                continue
            else:
                dests.add(dest)
        return dests

    @property
    def configs(self):
        return self._configs

    def add_config(self, config):
        self._configs.append(config)


def _all_parser_sections(parser):
    all_sections = {}
    for section in parser.sections():
        all_sections[section] = {}
        for option in parser.options(section):
            all_sections[section][option] = parser.get(section, option)
    return all_sections


def parse_file(filename):
    # Parse a file into a dict of name: options_dict
    # options_dict is a dict of option_name: value
    parser = StripQuotesConfigParser()
    sections = {}

    try:
        fname = parser.read(filename)
        if len(fname) == 0:
            logger.error("Unable to read configuration file %s", filename)
        else:
            sections = _all_parser_sections(parser)
    except MissingSectionHeaderError:
        logger.error("Configuration file %s contains no section headers", filename)
    except NoOptionError as e:
        logger.error(str(e))
    except InvalidPasswordFormat as e:
        logger.error(str(e))
    except InvalidOption as e:
        # When a configuration section has an Invalid Option, continue
        # See https://bugzilla.redhat.com/show_bug.cgi?id=1457101 for more info
        logger.warn("Invalid configuration detected: %s", str(e))
    except Exception as e:
        logger.error('Config file "%s" skipped because of an error: %s',
                     filename, str(e))
    return sections


# Helper methods used to validate parameters given to virt-who
def str_to_bool(value):
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.strip().lower() in ['yes', 'true', 'on', '1']:
            return True
        elif value.strip().lower() in ['no', 'false', 'off', '0']:
            return False
    raise ValueError("Unable to convert value to boolean")


def non_empty_string(value):
    if not isinstance(value, str):
        raise TypeError("Value is not a string")
    if not value:
        raise ValueError("String is empty")
    return value


def readable(path):
    if not os.access(path, os.R_OK):
        raise ValueError("Path '%s' is not accessible" % path)


def accessible_file(path):
    readable(path)
    if not os.path.isfile(path):
        raise ValueError("Path '%s' does not appear to be a file" % path)


def accessible_dir(path):
    readable(path)
    if not os.path.isdir(path):
        raise ValueError("Path '%s' does not appear to be a directory" % path)


def empty_or_accessible_files(paths):
    if not isinstance(paths, list):
        if not isinstance(paths, str):
            raise TypeError()
        if len(paths) == 0:
            return []
        paths = [paths]
    if len(paths) == 0:
        return paths
    for path in paths:
        accessible_file(path)
    return paths


class ValidationState(object):
    VALID = 'valid'
    INVALID = 'invalid'
    UNKNOWN = 'unknown'
    NEEDS_VALIDATION = 'needs_validation'


class ConfigSection(collections.MutableMapping):
    """
    This represents a section of configuration for virt-who. The interface that it exposes is
    dictionary like. This object maintains a state attribute. The state shows if the configuration
    section has passed validation, needs validation, or is invalid for some reason.
    """

    # The string representation of all default properties and values
    # Real values should be added in child classes
    DEFAULTS = ()

    __marker = object()

    def __init__(self, section_name, wrapper):
        """
        Initialization of ConfigSection instance
        :param section_name: The name of section
        :param wrapper: The parent of current config section
        """
        self.name = section_name
        self._wrapper = wrapper
        self.defaults = dict(self.DEFAULTS)
        # Those properties that have yet to be validated
        # Any item not in this set but in this ConfigSection is validated
        self._unvalidated_keys = set()
        self._invalid_keys = set()
        # Holds all properties for this section
        self._values = {}
        # Validation messages from last run
        self.validation_messages = []

        self.validation_methods = OrderedDict()
        self._destinations = {}
        self._required_keys = set()
        self._missing_required_keys = set()
        # Those values which should not be displayed
        self._restricted = set(['virt_type', 'sm_type'])

        # Add section defaults
        for key, value in self.defaults.items():
            self._values[key] = value
        self._update_state()

    def iteritems(self):
        for key in self:
            yield (key, self._values[key])

    def items(self):
        return util.DictItemsIter(self)

    def __iter__(self):
        return iter(self._values.keys())

    def _update_state(self):
        if len(self._unvalidated_keys) > 0:
            self.state = ValidationState.NEEDS_VALIDATION
        elif len(self._invalid_keys) > 0 or len(self._values) == 0 or len(
                self._missing_required_keys) > 0:
            self.state = ValidationState.INVALID
        else:
            self.state = ValidationState.VALID

    def __delitem__(self, key):
        if key in self:
            if self.has_default(key):
                return
            del self._values[key]
            if key in self._unvalidated_keys:
                self._unvalidated_keys.remove(key)
            if key in self._invalid_keys:
                self._invalid_keys.remove(key)
            self._update_state()
        else:
            raise KeyError('Unable to delete nonexistant property "%s"' % key)

    def __contains__(self, key):
        return key in self._values

    def __setitem__(self, key, value):
        if key not in self or (key in self and self[key] != value):
            self._unvalidated_keys.add(key)
            self._values[key] = value
            self._update_state()
        elif key not in self._unvalidated_keys:
            self._unvalidated_keys.add(key)
            self._update_state()

    def __len__(self):
        return len(self._values)

    def __getitem__(self, key):
        if key in self._values:
            return self._values[key]
        elif key in self.defaults:
            return self.defaults[key]
        else:
            raise KeyError('{0} not in {1}'.format(key, self.name))

    def _pre_validate(self):
        """
        Steps necessary to do before evaluation
        """
        if len(self._unvalidated_keys) == 0:
            if not self._values:
                self.validation_messages.append(('warning', 'No values provided in: %s' % self.name))
        else:
            for default_key in self.defaults.keys():
                if default_key not in self._unvalidated_keys and \
                   default_key not in self._restricted and \
                   self.defaults[default_key] is not None:
                    self.validation_messages.append(
                        (
                            'debug',
                            'Value for "%s" not set, using default: %s' %
                            (default_key, self.defaults[default_key])
                        )
                    )

    def check_required_keys(self):
        for required_key in self._required_keys:
            if required_key not in self and not self.has_default(required_key):
                msg = ('error', 'Required option: "%s" not set.'
                       % required_key)
                self.validation_messages.append(msg)
                self._missing_required_keys.add(required_key)

    def _post_validate(self):
        """
        Steps necessary to do after evaluation
        """
        self.check_required_keys()
        self.reset_to_defaults()
        self.apply_destinations()
        # Finally calls _update_state
        self._update_state()

    def _validate(self):
        if not self._unvalidated_keys:
            # Do not override validation_messages if there is nothing to validate
            return
        validation_messages = []

        # remove unknown keys
        unknown_keys = self._unvalidated_keys.difference(self.validation_methods.keys())
        self._unvalidated_keys.difference_update(unknown_keys)
        # Validate those keys that need to be validated
        for key, validation_method in self.validation_methods.items():
            if key not in self._unvalidated_keys:
                continue
            messages = validation_method(key)
            key_invalid = False
            if messages is not None:
                if isinstance(messages, list):
                    if any(message[0] == 'error' for message in messages):
                        key_invalid = True
                    validation_messages.extend(messages)
                else:
                    if messages[0] == 'error':
                        key_invalid = True
                    validation_messages.append(messages)
                # Only add if at least one result contains an 'error' level message
                if key_invalid:
                    self._invalid_keys.add(key)
            self._unvalidated_keys.remove(key)

        for key in unknown_keys:
            validation_messages.append(
                    ('warning', 'Ignoring unknown configuration option "%s"' % key)
            )
            if key in self._values:
                del self._values[key]

        self._update_state()
        self.validation_messages.extend(validation_messages)

    def validate(self):
        """
        Do validation of provided options.
        :return: list of validation messages
        """
        self._pre_validate()
        self._validate()
        self._post_validate()
        return self.validation_messages

    def reset_to_defaults(self):
        """
        When option is not set, then this methods tries to set such option to default value.
        """
        for key in self._invalid_keys.union(self.defaults.keys()):
            if self.has_default(key) and key not in self._values:
                self._values[key] = self.defaults[key]
                if key in self._required_keys and key not in self._restricted:
                    message = 'Required option: "%s" is missing in: "%s" using default "%s"' % \
                              (key, self.name, self.defaults[key])
                    self.validation_messages.append(('warning', message))

    def apply_destinations(self):
        # Rudimentary, should we be copying values around or just referencing them
        for key in self._destinations:
            if key in self._values and self._destinations[key] not in self._values and key \
                    not in self._invalid_keys:
                self._values[self._destinations[key]] = self._values[key]
                del self._values[key]

    def is_default(self, key):
        return key in self.defaults and (key not in self._values or self._values[key] == self.defaults[key])

    def is_section_default(self):
        """
        :return: This method returns True if this ConfigSection instance has exactly the same
        keys and values as the defined defaults (if any).
        :rtype: bool
        """
        return all(self.has_default(key) and self.is_default(key) for key in self)

    def has_default(self, key):
        return key in self.defaults

    def get(self, key, default=__marker):
        try:
            return self[key]
        except KeyError as e:
            if default is not self.__marker:
                return default
            raise e

    def __str__(self):
        return_string = "[%s]" % self.name
        for key, value in self.items():
            return_string += "\n%s=%s" % (key, value)
        return return_string

    def update(self, *args, **kwds):
        """
        This method implements update as usually defined on a regular dict. The only difference
        is a reference to self is expected.
        :param *args: Each arg passed if it has a "keys" method we will do the following d[k] =
            arg[k] for k in arg.keys. If not we treat the arg as iterable (possiblly a tuple of
            tuples or list of tuples etc). In this case we do the following: for key, val in arg:
            d[key] = value.
        :param **kwds: for each keyword arg in kwds we set d[keyword] = kwds[keyword]
        :return: Nothing
        """
        for arg in args:
            if getattr(arg, 'keys', None):
                for key in arg.keys():
                    self[key] = arg[key]
            else:
                for key, value in arg:
                    self[key] = value
        for key, value in kwds.items():
            self[key] = value

    def is_valid(self):
        return self.state == ValidationState.VALID

    @classmethod
    def get_defaults(cls):
        """
        Returns: A dictionary of the defaults defined for a ConfigSection of this type.
                 Strings are returned so as to match the values returned by parsers for other
                 sources of configuration (for example, ConfigParsers or argparse). Better to
                 treat args from all parsers as strings and to convert from strings
                 in one place than to check type multiple places up until then.
        """
        defaults = dict()
        for key, value in cls.DEFAULTS:
            defaults[key] = str(value)
        return defaults

    @classmethod
    def class_for_type(cls, virt_type, parent=None):
        if parent is None:
            clazz = cls
        else:
            clazz = parent
        for subclass in cls.__subclasses__():
            if getattr(subclass, 'VIRT_TYPE', None) == virt_type:
                clazz = subclass
                break
            # When VIRT_TYPE was not found in this subclass, then
            # this subclass can have another subclasses :-)
            else:
                result = subclass.class_for_type(virt_type, cls)
                if result != cls:
                    clazz = result
                    break
        return clazz

    @classmethod
    def from_dict(cls, values, section_name, wrapper):
        virt_type = values.get('virt_type', None) or values.get('type')
        sub_cls = cls.class_for_type(virt_type)
        section = sub_cls(section_name, wrapper)
        section.update(**values)
        return section

    def _validate_str_to_bool(self, key):
        result = None
        try:
            self._values[key] = str_to_bool(self._values[key])
        except (KeyError, ValueError):
            if self.has_default(key):
                self._values[key] = str_to_bool(self.defaults[key])
                result = (
                    'warning',
                    '%s must be a valid boolean, using default. '
                    'See man virt-who-config for more info' % key
                )
            else:
                result = (
                    'warning',
                    '%s must be a valid boolean, ignoring. '
                    'See man virt-who-config for more info' % key
                )
        return result

    def _validate_non_empty_string(self, key):
        result = None
        try:
            value = self._values[key]
        except KeyError:
            if not self.has_default(key):
                result = ('warning', 'Value for %s not set' % key)
        else:
            if not isinstance(value, str):
                result = ('warning', '%s is not set to a valid string, using default' % key)
            elif len(value) == 0:
                result = ('warning', '%s cannot be empty, using default' % key)
        return result

    def _validate_list(self, list_key):
        result = None
        if self.is_default(list_key):
            self._values[list_key] = []
        elif isinstance(self[list_key], list):
            filtered_items = [item for item in self._values[list_key] if isinstance(item, str)]
            self._values[list_key] = filtered_items
        elif isinstance(self._values[list_key], str):
            self._values[list_key] = parse_list(self._values[list_key])
        else:
            result = (
                'warning',
                'Option "%s" must be one or more strings, ignoring' % list_key
            )
            self._values[list_key] = []  # Reset to empty list
        return result

    def add_key(self, key, validation_method=None, default=__marker, destination=None,
                required=False, restricted=False):
        if default is not self.__marker:
            self.defaults[key] = default
        if not validation_method:
            raise AttributeError('validation_method must be provided for {0}'.format(key))
        self.validation_methods[key] = validation_method
        if destination:
            self._destinations[key] = destination

        if required:
            self._required_keys.add(key)
        elif key in self._required_keys:
            self._required_keys.remove(key)

        if restricted:
            self._restricted.add(key)
        elif key in self._restricted:
            self._restricted.remove(key)

    def remove_key(self, key):
        # The parts of the ConfigSection that might contain "key"
        trackers = [self.validation_methods,
                    self._destinations,
                    self._restricted,
                    self._required_keys,
                    self.defaults,
                    self._values,
                    self._unvalidated_keys,
                    ]
        for tracker in trackers:
            if key in tracker:
                if isinstance(tracker, collections.MutableMapping):
                    del tracker[key]
                elif isinstance(tracker, collections.MutableSequence):
                    tracker.remove(key)
                elif isinstance(tracker, collections.MutableSet):
                    tracker.discard(key)


class VirtConfigSection(ConfigSection):
    """
    This class is used for validation of virtualization backend section.
    It tries o validate options that are common for all virtualization
    backends supported by virt-who.
    """

    DEFAULTS = (
        ('sm_type', 'sam'),
    )
    RENAMED_OPTIONS = (
        ('filter_host_uuids', 'filter_hosts'),
        ('exclude_host_uuids', 'exclude_hosts'),
    )
    HYPERVISOR_ID = ()

    def __init__(self, section_name, wrapper):
        super(VirtConfigSection, self).__init__(section_name, wrapper)
        self.add_key('type', validation_method=self._validate_virt_type, default='libvirt')
        self.add_key('sm_type', validation_method=self._validate_sm_type, default="sam", restricted=True)
        self.add_key('hypervisor_id', validation_method=self._validate_hypervisor_id, default='uuid')
        # Unencrypted passwords
        self.add_key('password', validation_method=self._validate_unencrypted_password)
        self.add_key('rhsm_password', validation_method=self._validate_unencrypted_password)
        self.add_key('rhsm_proxy_password', validation_method=self._validate_unencrypted_password)
        self.add_key('sat_password', validation_method=self._validate_unencrypted_password)
        # Encrypted passwords
        self.add_key('encrypted_password', validation_method=self._validate_encrypted_password,
                     destination='password')
        self.add_key('rhsm_encrypted_password', validation_method=self._validate_encrypted_password,
                     destination='rhsm_password')
        self.add_key('rhsm_encrypted_proxy_password', validation_method=self._validate_encrypted_password,
                     destination='rhsm_proxy_password')
        self.add_key('sat_encrypted_password', validation_method=self._validate_encrypted_password,
                     destination='sat_password')
        # Usernames
        self.add_key('username', validation_method=self._validate_username)
        self.add_key('rhsm_username', validation_method=self._validate_username)
        self.add_key('rhsm_proxy_user', validation_method=self._validate_username)
        self.add_key('sat_username', validation_method=self._validate_username)
        # Needed to allow us to parse the destination info
        self.add_key('sat_server', validation_method=lambda *args: None)
        self.add_key('server', validation_method=self._validate_server)
        self.add_key('owner', validation_method=self._validate_owner, required=True)
        self.add_key('filter_hosts', validation_method=self._validate_filter)
        self.add_key('filter_type', validation_method=self._validate_filter_type)
        self.add_key('exclude_hosts', validation_method=self._validate_filter)
        self.add_key('rhsm_proxy_hostname', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_proxy_port', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_hostname', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_port', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_prefix', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_insecure', validation_method=self._validate_non_empty_string)
        self.add_key('rhsm_no_proxy', validation_method=self._validate_non_empty_string)

    def __setitem__(self, key, value):
        for old_key, new_key in self.RENAMED_OPTIONS:
            if key == old_key:
                key = new_key
        super(VirtConfigSection, self).__setitem__(key, value)

    def _pre_validate(self):
        # Used to allow virt_type to be provided for type (if type was not given)
        if 'virt_type' in self._values and 'type' not in self._values:
            self['type'] = self._values['virt_type']
            self.remove_key('virt_type')
        # Logic of owner is little bit tricky :-(
        if 'sm_type' in self._values:
            if self._values['sm_type'] == SAT5:
                self._required_keys.update(Satellite5DestinationInfo.required_kwargs)
                # Owner is only necessary for SAT6 and only for certain backends
                self._required_keys.discard('owner')
            elif self._values['sm_type'] == SAT6:
                if self['type'] == 'libvirt' and 'server' not in self._values:
                    # Owner is not necessary sam an libvirt virt backend
                    self._required_keys.discard('owner')
                elif self['type'] == 'vdsm':
                    self._required_keys.discard('owner')
                # TODO: is_hypervisor has to be true too, but it is hard to implement in this state
                elif self['type'] == 'fake':
                    self._required_keys.discard('owner')
        super(VirtConfigSection, self)._pre_validate()

    def _validate_sm_type(self, key):
        result = None
        try:
            sm_type = non_empty_string(self._values['sm_type'])
        except KeyError:
            result = ('error', 'Unknown destination type')
        except (ValueError, TypeError) as e:
            result = ('error', str(e))
        else:
            acceptable_sm_types = [SAT5, SAT6]
            if sm_type not in acceptable_sm_types:
                result = ('error', '%s must be one of %s' % (key, ', '.join(acceptable_sm_types)))
        return result

    def _validate_virt_type(self, key):
        result = None
        try:
            virt_type = self._values[key]
        except KeyError:
            result = ('warning', 'Virt. type is not set, using default')
        else:
            if virt_type not in VW_TYPES:
                result = ('warning', 'Unsupported virt. type is set, using default')
                self._values[key] = self.defaults[key]
        return result

    # def _validate_hypervisor_id(self, key='hypervisor_id'):
    #     # Note: Every virt. backend should implement own validation method of hypervisor id
    #     return None

    def _validate_hypervisor_id(self, key='hypervisor_id'):
        """
        Do validation of hypervisor_id, when virtualization backend support it
        (self.HYPERVISOR_ID contains at leas one string)
        :param key: hypervisor ID
        :return: None or warning, when unsupported hypervisor_id is provided.
        """
        result = []
        if key in self._values and len(self.HYPERVISOR_ID) > 0:
            if self._values[key] not in self.HYPERVISOR_ID:
                result.append((
                    'error',
                    'Invalid option: "%s" for hypervisor_id, use one of: (%s)' %
                    (self._values[key], ", ".join(self.HYPERVISOR_ID))
                ))
        if len(result) > 0:
            return result
        else:
            return None

    def _validate_unencrypted_password(self, pass_key):
        """
        Try to validate unencrypted password. It has to be UTF-8 encoded.
        :param pass_key: This could be: 'password', 'rhsm_password',
                         'rhsm_proxy_password' and 'sat_password'
        """
        result = None
        try:
            password = self._values[pass_key]
        except KeyError:
            result = (
                'warning',
                'Option: "%s" was not set in configuration: %s' % (pass_key, self.name)
            )
        else:
            if password != NotSetSentinel:
                try:
                    if (not six.PY3 and isinstance(password, str)) or isinstance(password, bytes):
                        password.decode('utf-8')
                except UnicodeDecodeError:
                    result = (
                        'warning',
                        "Value: {0} of option '{1}': is not in UTF-8 encoding".format(password, pass_key)
                    )
        return result

    def _validate_encrypted_password(self, pass_key):
        """
        Try to validate encrypted password. It has to be UTF-8 encoded.
        :param pass_key: This could be: 'encrypted_password', 'rhsm_encrypted_password',
                         'rhsm_encrypted_proxy_password' and 'sat_encrypted_password'
        """
        result = None
        decrypted_pass_key = self._destinations[pass_key]
        try:
            pwd = self._values[pass_key]
        except KeyError:
            result = (
                'warning',
                'Option: "%s" was not set in configuration %s' % (pass_key, self.name)
            )
        else:
            try:
                self._values[decrypted_pass_key] = Password.decrypt(unhexlify(pwd))
            except (TypeError, IndexError, UnicodeDecodeError, BinasciiError):
                result = (
                    'warning',
                    "Option \"{option}\" cannot be decrypted, possibly corrupted"
                    .format(option=pass_key)
                )
            except InvalidKeyFile as err:
                result = (
                    'error',
                    "Unable to read key file: %s" % str(err)
                )
        return result

    def _validate_username(self, username_key):
        """
        Try to validate username
        :param username_key: Possible values could be: 'username', 'rhsm_username',
                             'rhsm_proxy_username', 'sat_username'
        """
        result = None
        try:
            username = self._values[username_key]
        except KeyError:
            result = ('warning', 'Option: "%s" was not set in configuration: %s' % (username_key, self.name))
        else:
            if username != NotSetSentinel:
                try:
                    username.encode('latin1')
                except (UnicodeEncodeError, UnicodeDecodeError):
                    result = (
                        'warning',
                        "Value: {0} of option '{1}': is not in latin1 encoding".format(
                            username,
                            username_key
                        )
                    )
        return result

    def _validate_server(self, key):
        """
        Try to validate server definition
        """
        result = None
        # Server option must be there for ESX, RHEVM, and HYPERV
        if key not in self._values or len(self._values[key]) == 0:
            if 'type' in self._values and self._values['type'] in ['libvirt', 'vdsm', 'fake']:
                self._values[key] = ''
            else:
                result = (
                    'error',
                    "Option %s needs to be set in config: '%s'" % (key, self.name)
                )

        return result

    def _validate_owner(self, key):
        """
        Try to validate environment option
        """
        result = None
        sm_type = self._values['sm_type']
        virt_type = self._values.get('type')
        if sm_type == 'sam' and virt_type in ('esx', 'rhevm', 'hyperv', 'xen', 'ahv'):
            if key not in self:
                result = (
                    'warning',
                    "Option `%s` needs to be set in config: '%s'" % (key, self.name)
                )
        return result

    def _validate_filter_type(self, key):
        """
        Try to validate type of filter
        :param key: key of filter type (it should be always 'filter_type')
        :return: None or list of warnings
        """
        result = None
        filter_type = self._values.get('filter_type')
        if filter_type not in FILTER_TYPES:
            result = (
                'error',
                "'%s' must be one of: '%s'" % (key, ", ".join(FILTER_TYPES))
            )
        return result

    def _validate_filter(self, filter_key):
        """
        Try to validate filter option. It can contain list of hostname and UUIDs
        :param filter_key: key of filter ('filter_hosts', 'exclude_hosts')
        :return: None or list of warnings
        """

        # Check validity of list first
        result = self._validate_list(filter_key)
        if result is not None:
            # When validation of list failed, then there is no reason for
            # further validation, because self._values[filter_key] is empty
            return result

        result = []

        hypervisor_id = self._values.get('hypervisor_id')

        # When hypervisor_id is not 'uuid' (it is hostname or hwuuid), then
        # filter should not contain any UUID, because no UUIDs will be gathered
        # Note: example of hwuuid is e.g. 'host-9' (it doesn't look like UUID)
        if hypervisor_id != 'uuid':
            wrong_filter_values = []
            for filter_value in self._values[filter_key]:
                try:
                    uuid.UUID(filter_value)
                except ValueError:
                    pass
                else:
                    wrong_filter_values.append(filter_value)
            if len(wrong_filter_values) > 0:
                result.append((
                    'warning',
                    'Filter values: "%s" appear to be UUIDs. UUIDs are not gathered when hypervisor_id = "%s"' %
                    (', '.join(wrong_filter_values), hypervisor_id)
                ))

        if 'filter_type' not in self:
            result.append((
                'warning',
                '"%s" is set, but filter_type is not set. Possible values of filter_type: "%s". Using default: "wildcards".' %
                (filter_key, ', '.join(FILTER_TYPES))
            ))

        if len(result) == 0:
            result = None

        return result


class GlobalSection(ConfigSection):
    """
    Class used for validation of global section
    """
    SECTION_NAME = VW_GLOBAL

    def __init__(self, *args, **kwargs):
        super(GlobalSection, self).__init__(*args, **kwargs)
        self.add_key('debug', validation_method=self._validate_str_to_bool, default=False)
        self.add_key('oneshot', validation_method=self._validate_str_to_bool, default=False)
        self.add_key('print_', validation_method=self._validate_str_to_bool, default=False, destination='print')
        self.add_key('log_per_config', validation_method=self._validate_str_to_bool, default=False)
        self.add_key('configs', validation_method=self._validate_list, default=[])
        self.add_key('reporter_id', validation_method=self._validate_non_empty_string,
                     default=util.generateReporterId())
        self.add_key('interval',  validation_method=self._validate_interval, default=DefaultInterval)
        self.add_key('log_file', validation_method=self._validate_non_empty_string, default=log.DEFAULT_LOG_FILE)
        self.add_key('log_dir', validation_method=self._validate_non_empty_string, default=log.DEFAULT_LOG_DIR)

    def _validate_interval(self, key):
        result = None
        try:
            self._values[key] = int(self._values[key])
        except (TypeError, ValueError) as e:
            self._values[key] = 0
        except KeyError:
            return ('warning', '%s is missing' % key)

        if self._values[key] < MinimumSendInterval:
            message = "Interval value can't be lower than {min} seconds. Default value of " \
                      "{default} " \
                      "seconds will be used.".format(min=MinimumSendInterval, default=DefaultInterval)
            result = ("warning", message)
            self._values['interval'] = DefaultInterval
        return result

    def _validate_configs(self):
        return self._validate_list('configs')

    def _validate(self):
        super(GlobalSection, self)._validate()

        if self.get('print_', None) or self.get('print', None):
            self._values['oneshot'] = True

# String representations of all the default configuration for virt-who
DEFAULTS = {
    VW_GLOBAL: GlobalSection.get_defaults(),
    VW_ENV_CLI_SECTION_NAME: {
        'sm_type': 'sam',
        'virt_type': 'libvirt',
    },
}


class EffectiveConfig(collections.MutableMapping):
    """
    This object represents the total configuration of virt-who including all global parameters
    and all sections that define a source or destination.
    """

    __marker = object()

    def __iter__(self):
        return iter(self._sections)

    def __delitem__(self, key):
        if key in self:
            del self._sections[key]
        else:
            raise KeyError("Unable to delete nonexistant section '%s'" % key)

    def __setitem__(self, key, value):
        self._sections[key] = value

    def __len__(self):
        return len(self._sections)

    def __getitem__(self, key):
        return self._sections[key]

    def __contains__(self, item):
        return item in self._sections

    def items(self):
        return util.DictItemsIter(self)

    def __init__(self):
        self.validation_messages = []
        self._sections = {}

    def validate(self):
        validation_messages = []
        for section_name, section in self._sections.items():
            validation_messages.extend(
                    [(level, "[%s]: %s" % (section.name, message))
                     for (level, message) in section.validate()]
            )
        self.validation_messages = validation_messages
        return validation_messages

    def is_valid(self):
        return all(child.state == ValidationState.VALID for (name, child) in self.items())

    @staticmethod
    def filter_parameters(desired_parameters, values_to_filter):
        matching_parameters = {}
        non_matching_parameters = {}
        for param, value in values_to_filter.items():
            if value is None:
                continue
            if isinstance(value, list):
                value = [str(item) for item in value]
            else:
                value = str(value)
            if param in desired_parameters:
                matching_parameters[param] = value
            else:
                # Take all parameters and their values for those params that are non-default or
                # for which a default is unknown
                non_matching_parameters[param] = value
        return matching_parameters, non_matching_parameters

    @staticmethod
    def all_drop_dir_config_sections(config_dir=VW_CONF_DIR):
        """
        Read all configuration sections in the default config directory
        :return: a dictionary of {name: {key: value, ... } ... }
        """
        try:
            all_dir_content = set(os.listdir(config_dir))
            conf_files = set(s for s in all_dir_content if s.endswith('.conf'))
            non_conf_files = all_dir_content - conf_files
        except OSError:
            logger.warn("Configuration directory '%s' doesn't exist or is not accessible",
                        config_dir)
            return {}

        if not all_dir_content:
            logger.warn("Configuration directory '%s' appears empty", config_dir)
        elif not conf_files:
            logger.warn("Configuration directory '%s' does not have any '*.conf' files but "
                        "is not empty", config_dir)
        elif non_conf_files:
            logger.debug("There are files in '%s' not ending in '*.conf' is this "
                         "intentional?", config_dir)
        all_sections = {}
        for conf in conf_files:
            if conf.startswith('.'):
                continue
            all_sections.update(parse_file(os.path.join(config_dir, conf)))
        return all_sections

    def is_default(self, section, option):
        return self._sections[section].is_default(option)

    def get(self, section, default=__marker):
        return self._sections[section]

    def virt_sections(self):
        """
        :return: list of sections that represent virt_backends
        """
        return [(name, section) for (name, section) in self.items()
                if name not in [VW_GLOBAL, VW_VIRT_DEFAULTS_SECTION_NAME]]


def _check_effective_config_validity(effective_config):
    validation_errors = effective_config.validate()
    valid_virt_sections = {}
    invalid_virt_sections = {}
    for name, section in effective_config.virt_sections():
        if section.is_valid():
            valid_virt_sections[name] = section
        else:
            invalid_virt_sections[name] = section
    has_non_default_env_cli = False

    # Drop invalid configurations first
    if len(invalid_virt_sections.keys()) > 0:
        for name, section in invalid_virt_sections.items():
            if name == VW_ENV_CLI_SECTION_NAME and not section.is_section_default():
                has_non_default_env_cli = True
            validation_errors.append(('warning', 'Dropping invalid configuration "%s"' % name))
            del effective_config[name]

        # If there are no valid virt sections
        if not valid_virt_sections:
            validation_errors.append(('warning', 'No valid configurations found'))

    # In order to keep compatibility with older releases of virt-who,
    # fallback to using libvirt as default virt backend
    # only if we did not have a non_default env/cmdline config
    if not has_non_default_env_cli and len(effective_config.virt_sections()) == 0 and len(
            invalid_virt_sections) == 0:
        effective_config[VW_ENV_CLI_SECTION_NAME] = ConfigSection.from_dict(
                DEFAULTS[VW_ENV_CLI_SECTION_NAME],
                VW_ENV_CLI_SECTION_NAME,
                effective_config)
        validation_errors.append(('warning',
                                  'Using default "%s" configuration' % VW_ENV_CLI_SECTION_NAME))

    effective_config.validate()
    return effective_config, validation_errors


def init_config(env_options, cli_options, config_dir=None):
    """
    Initialize and return the effective virt-who configuration
    :param env_options: The dict of options parsed from the environment
    :param cli_options: The dict of options parsed from the CLI
    :param config_dir: The path to directory containing configuration files
    :return: EffectiveConfig
    """

    if config_dir is None:
        config_dir = VW_CONF_DIR

    validation_errors = []
    effective_config = EffectiveConfig()
    global logger  # Use module level logger as this is likely called before other logging init

    effective_config[VW_GLOBAL] = GlobalSection(VW_GLOBAL, effective_config)
    effective_config[VW_ENV_CLI_SECTION_NAME] = ConfigSection(VW_ENV_CLI_SECTION_NAME,
                                                              effective_config)
    effective_config[VW_ENV_CLI_SECTION_NAME].defaults = DEFAULTS[VW_ENV_CLI_SECTION_NAME]

    global_required_params = list(effective_config[VW_GLOBAL].defaults.keys())

    # Split environment variables values into global or non
    env_globals, env_non_globals = effective_config.filter_parameters(global_required_params,
                                                                      env_options)
    # Split cli variables values into global or non
    cli_globals, cli_non_globals = effective_config.filter_parameters(global_required_params,
                                                                      cli_options)
    # Read the virt-who general conf file
    vw_conf = parse_file(VW_GENERAL_CONF_PATH)

    global_section = vw_conf.pop(VW_GLOBAL, {})
    # NOTE: Might be nice in the future to include the defaults in this object
    # So that section would still exist in the output
    virt_defaults_section = vw_conf.pop(VW_VIRT_DEFAULTS_SECTION_NAME, {})
    global_section_sources = [global_section, env_globals, cli_globals]

    for global_source in global_section_sources:
        for key, value in global_source.items():
            if key:
                effective_config[VW_GLOBAL][key.lower()] = value

    # Validate GlobalSection before use.
    effective_config[VW_GLOBAL].validate()

    # Close existing logger
    log.closeLogger(logger)
    # Initialize logger, the creation of this object is the earliest it can be done
    log.init(effective_config)
    logger = log.getLogger(config=effective_config, queue=False)

    # Create the effective env / cli config from those values we've sorted out as non_global
    env_cli_sources = [env_non_globals, cli_non_globals]
    for env_cli_source in env_cli_sources:
        for key, value in env_cli_source.items():
            if key:
                effective_config[VW_ENV_CLI_SECTION_NAME][key.lower()] = value

    # 1638250: issue in the urllib or requests package in python 3
    if six.PY3:
        https = None
        if 'https_proxy' in os.environ:
            https = os.environ['https_proxy']
        http = None
        if 'http_proxy' in os.environ:
            http =  os.environ['http_proxy']
            del os.environ['http_proxy']
        if http and not https:
            os.environ['https_proxy'] = http

    # Now with the aggregate config data, run it through the appropriate class to get it validated/defaulted.
    effective_config[VW_ENV_CLI_SECTION_NAME] = VirtConfigSection.from_dict(effective_config[VW_ENV_CLI_SECTION_NAME], VW_ENV_CLI_SECTION_NAME, effective_config)

    if effective_config[VW_ENV_CLI_SECTION_NAME].is_section_default():
        del effective_config[VW_ENV_CLI_SECTION_NAME]

    all_sections_to_add = {}

    # read the files provided to the configs var if defined
    for file_name in effective_config[VW_GLOBAL]['configs']:
        logger.info("Using configuration passed in by -c/--configs; ignoring configuration files in '%s'", config_dir)
        all_sections_to_add.update(parse_file(filename=file_name))

    if len(all_sections_to_add) == 0:
        # read sections from /etc/virt-who.conf
        all_sections_to_add.update(vw_conf)
        # also read all sections in conf files in the drop dir
        all_sections_to_add.update(effective_config.all_drop_dir_config_sections(config_dir=config_dir))

    # We should ignore any additional sections defined as VW_GLOBAL
    # or VW_VIRT_DEFAULTS_SECTION_NAME as we've already parsed those
    if VW_GLOBAL in all_sections_to_add:
        del all_sections_to_add[VW_GLOBAL]
    if VW_VIRT_DEFAULTS_SECTION_NAME in all_sections_to_add:
        del all_sections_to_add[VW_VIRT_DEFAULTS_SECTION_NAME]

    for section, values in all_sections_to_add.items():
        new_section = {}
        new_section.update(**virt_defaults_section)
        new_section.update(**values)
        try:
            new_section = ConfigSection.from_dict(new_section, section, effective_config)
        except KeyError:
            # Missing required attribute
            continue
        effective_config[section] = new_section

    effective_config, errors = _check_effective_config_validity(effective_config)
    validation_errors.extend(errors)

    # Log pending errors
    for err in validation_errors:
        method = getattr(logger, err[0])
        if method is not None and err[0] in ['error', 'warning', 'info', 'debug']:
            method(err[1])

    return effective_config



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
# Module for STOMP messaging, part of virt-who
#
# Copyright (C) 2018 Kevin Howell <khowell@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Implements enough of the STOMP specification (https://stomp.github.io/stomp-specification-1.2.html) to enable
implementation of a client for VDSM's JSON-RPC API over STOMP. Could be easily improved to be a general STOMP client;
TODOs and FIXMEs document known gaps.

There are better developed STOMP libraries for Python available (such as stomper), but for RHEL/CentOS, at the time of
writing, only available in EPEL.
"""

import logging
import socket
from six import text_type
import uuid

log = logging.getLogger(__name__)


class StompFrame(object):
    """STOMP frame as described in the STOMP specification."""

    def __init__(self, command, headers=None, body=None):
        """Create a new StompFrame.

        :param command: STOMP command as a string (ex. 'CONNECT').
        :param headers: dict containing STOMP headers.
        :param body: body as bytes.
        """
        self.command = command
        self.headers = headers or {}
        self.body = body or b''

    @classmethod
    def from_bytes(cls, data):
        """Decode bytes into a StompFrame.

        :param data: a single STOMP frame as bytes.
        :return: a StompFrame.
        :raises:
        """
        parts = [part for part in data.splitlines()]
        if not parts:
            raise IOError('Cannot parse data as STOMP frame')

        # find indexOf empty line, this is the indicator what follows is body
        try:
            end_of_headers_index = parts.index(b'')
        except ValueError:
            raise IOError('Cannot parse data as STOMP frame')

        command_part = parts[0]
        header_parts = parts[1:end_of_headers_index]
        body_parts = parts[end_of_headers_index:]  # TODO assert that remaining data ends in EOF (NULL)

        return StompFrame(
            command=command_part.decode('utf-8'),
            headers=cls._decode_headers(header_parts),
            body=cls._unescape_bytes(b''.join(body_parts)[:-1]),  # TODO handle content-length
        )

    def to_bytes(self):
        """Encode the STOMP frame as bytes.

        :return: bytes.
        """
        frame = b'\n'.join([
            self.command.encode('utf-8'),
            self._encode_headers(),
            self.body
        ])
        frame = frame + b'\x00'
        return frame

    def _encode_headers(self):
        if not self.headers:
            return b''
        return b'\n'.join(
            [b':'.join(
                [
                    self._escape_bytes(key.encode('utf-8')),
                    self._escape_bytes(text_type(value).encode('utf-8')),
                ]
            ) for key, value in sorted(self.headers.items())]  # sort the headers for predictable messages
        ) + b'\n'

    @classmethod
    def _decode_headers(cls, data):
        headers = {}
        for line in data:
            key, value = line.split(b':')  # TODO err intelligently if more than one ':' detected
            key = cls._unescape_bytes(key).decode('utf-8')
            value = cls._unescape_bytes(value).decode('utf-8')
            if key not in headers:  # first value wins
                headers[key] = value
        return headers

    @staticmethod
    def _escape_bytes(data):
        if data is None:
            return b''
        data = data.replace(b'\\', b'\\\\')  # \ -> \\
        data = data.replace(b'\r', b'\\r')   # CR -> \r
        data = data.replace(b'\n', b'\\n')   # LF -> \n
        data = data.replace(b':', b'\\c')    # : -> \c
        return data

    @staticmethod
    def _unescape_bytes(data):
        data = data.replace(b'\\c', b':')    # \c -> :
        data = data.replace(b'\\n', b'\n')   # \n -> LF
        data = data.replace(b'\\r', b'\r')   # \r -> CR
        # TODO could check here that there are illegal escape sequences (ex. \t), see STOMP spec for details
        data = data.replace(b'\\\\', b'\\')  # \\ -> \
        return data


class StompClient(object):
    """Minimal STOMP client with SSL support."""

    def __init__(self, host, port, ssl_context=None, timeout=None):
        """Create a new STOMP client.

        :param host: STOMP server hostname.
        :param port: STOMP server port.
        :param ssl_context: SSL context if using TLS.
        :param timeout: timeout for blocking operations.
        """
        self.host = host
        self.port = port
        self.ssl_context = ssl_context
        self.timeout = timeout
        self.socket = None
        self.buffer = b''  # FIXME for performance, could construct a buffer and then recv_into instead...
        self.subscription_ids = []

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def connect(self):
        """Connect to the STOMP server.

        Sends a STOMP CONNECT and consumes the subsequent CONNECTED message.

        :raises: IOError if no connection attempts succeed.
        """
        for family, _, _, _, addr in socket.getaddrinfo(self.host, self.port):
            try:
                self.socket = socket.socket(family)
                self.socket.settimeout(self.timeout)
                if self.ssl_context:
                    if hasattr(self.ssl_context, 'wrap_socket'):
                        self.socket = self.ssl_context.wrap_socket(self.socket)
                        self.socket.connect(addr)
                    else:
                        import M2Crypto.SSL.Connection
                        self.socket = M2Crypto.SSL.Connection(self.ssl_context)
                        self.socket.connect((self.host, int(self.port)))
                message = StompFrame('CONNECT', headers={
                    'accept-version': '1.2',
                    'host': self.host,
                })
                self.socket.sendall(message.to_bytes())
                self._validate_connected(self._recv_frame())
                return  # success connecting
            except (socket.error, IOError) as e:
                log.warning('Unable to connect %s:%s: %s' % (self.host, self.port, text_type(e)))
        raise IOError('Unable to connect to %s:%s' % (self.host, self.port))

    @staticmethod
    def _validate_connected(data):
        frame = StompFrame.from_bytes(data)
        if frame.command != 'CONNECTED':
            raise IOError('Did not parse expected CONNECTED frame')

    def close(self):
        """Close STOMP server connection, unsubscribing from any tracked subscriptions."""
        for subscription_id in self.subscription_ids:
            self.unsubscribe(subscription_id)
        self.socket.close()
        self.socket = None

    def send(self, command, headers=None, data=None):
        """Construct and send a STOMP frame.

        :param command: STOMP command as a string.
        :param headers: dict containing STOMP headers.
        :param data: STOMP body as bytes.
        """
        self.socket.sendall(StompFrame(command, headers, data).to_bytes())

    def receive(self):
        """Receive a STOMP frame.

        :return: a StompFrame.
        """
        return StompFrame.from_bytes(self._recv_frame())

    def _recv_frame(self):
        while b'\x00' not in self.buffer:
            self.buffer += self.socket.recv(4096)
        frame, self.buffer = self.buffer.split(b'\x00', 1)
        # add null byte back to frame
        frame += b'\x00'
        return frame

    def subscribe(self, destination):
        """Subscribe to a queue or topic.

        :param destination: Name of the queue or topic to subscribe to.
        :return: randomly generated subscription ID.
        """
        subscription_id = uuid.uuid4()
        self.send('SUBSCRIBE', {
            'id': subscription_id,
            'destination': destination,
        })
        self.subscription_ids.append(subscription_id)
        return subscription_id

    def unsubscribe(self, subscription_id):
        """Unsubscribe from a given subscription ID.

        :param subscription_id: id returned by subscribe call.
        """
        self.send('UNSUBSCRIBE', {
            'id': subscription_id,
        })



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
# Module for JSON-RPC over STOMP, part of virt-who
#
# Copyright (C) 2018 Kevin Howell <khowell@redhat.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Minimal implementation of the JSON-RPC specification (http://www.jsonrpc.org/specification) over STOMP to enable a
client for VDSM.

There are better developed JSON-RPC libraries for Python available (such as jsonrpclib), but for RHEL/CentOS, at the
time of writing, only available in EPEL.
"""

import json
import ssl
from six import text_type
import uuid

from virtwho.virt.vdsm.stomp import StompClient


class JsonRpcClient(object):
    """JSON-RPC 2.0 over STOMP client"""

    def __init__(self, host, port, ssl_context=None, timeout=None):
        """Create a JSON-RPC over STOMP client.

        :param host: STOMP server hostname.
        :param port: STOMP server port.
        :param ssl_context: SSL context if using TLS or None.
        :param timeout: timeout for blocking operations.
        """
        self.id = text_type(uuid.uuid4())
        self.stomp = StompClient(host, port, ssl_context=ssl_context, timeout=timeout)

    def connect(self):
        """Connect to the STOMP server."""
        self.stomp.connect()
        self.stomp.subscribe(self.id)

    def close(self):
        """Close the underlying connection."""
        self.stomp.close()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def call(self, method, **params):
        """Call a method via JSON-RPC.

        :param method: name of method.
        :param params: parameters for the method.
        :return: the result as a dict if successful.
        :raises: RuntimeError if the result is an error or cannot be decoded.
        """
        data_object = {
            'jsonrpc': '2.0',
            'method': method,
            'id': text_type(uuid.uuid4())
        }
        if params:
            data_object['params'] = {}
            for key, value in params.items():
                data_object['params'][key] = value
        data = json.dumps(data_object).encode('utf-8')
        self.stomp.send('SEND', {
            'destination': 'jms.topic.vdsm_requests',
            'content-length': len(data),
            'reply-to': self.id,
        }, data)
        response = self.stomp.receive()
        if response.command == 'ERROR':
            raise RuntimeError(response.body.decode('utf-8'))
        response_json = json.loads(response.body.decode('utf-8'))
        if 'error' in response_json:
            # TODO log verbose error object?
            raise RuntimeError(response_json['error']['message'])
        return response_json['result']


def main():
    """Utility for one-off calls to VDSM via JSON-RPC over STOMP.

    Example invocation:

    python -m virtwho.virt.vdsm.jsonrpc
        --cert /etc/pki/vdsm/certs/vdsmcert.pem
        --key /etc/pki/vdsm/keys/vdsmkey.pem
        $HOSTNAME 54321
        Host.getVMList onlyUUID=False
    """
    import argparse

    parser = argparse.ArgumentParser(description='vdsm over jsonrpc utility')

    parser.add_argument('--key', help='path to key pem file')
    parser.add_argument('--cert', help='path to cert pem file')
    parser.add_argument('host', help='hostname')
    parser.add_argument('port', help='port')
    parser.add_argument('method', help='method to call via JSON-RPC')
    parser.add_argument('args', help='arguments for the call with format $key=$value', nargs='*')
    args = parser.parse_args()

    ssl_enabled = 'cert' in args and 'key' in args
    if ('key' in args or 'cert' in args) and not ssl_enabled:
        raise ValueError('Must provide both --cert and --key for SSL')

    ssl_context = None
    if ssl_enabled:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        ssl_context.verify_mode = ssl.CERT_NONE
        ssl_context.load_cert_chain(args.cert, args.key)

    # parse method args
    method_args = {}
    for arg in args.args:
        key, value = arg.split('=')
        key = key.strip()
        value = value.strip()
        if value.lower() == 'true':
            value = True
        if value.lower() == 'false':
            value = False
        method_args[key] = value

    with JsonRpcClient(args.host, args.port, ssl_context=ssl_context) as jsonrpc_client:
        print(json.dumps(jsonrpc_client.call(args.method, **method_args), indent=True))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import print_function, absolute_import


from .virt import (Virt, VirtError, Guest, AbstractVirtReport, DomainListReport,
                  HostGuestAssociationReport, ErrorReport,
                  Hypervisor, DestinationThread, IntervalThread, info_to_destination_class)

__all__ = ['Virt', 'VirtError', 'Guest', 'AbstractVirtReport',
           'DomainListReport', 'HostGuestAssociationReport',
           'ErrorReport', 'Hypervisor', 'DestinationThread',
           'IntervalThread', 'info_to_destination_class']



from __future__ import print_function
from mock import patch, MagicMock, PropertyMock

from base import TestBase

from virtwho.util import RequestsXmlrpcTransport


class FakeParser(object):
    def feed(self, arg):
        self.arg = arg

    def close(self):
        pass


class TestRequestsXmlrpcTransport(TestBase):
    @patch('requests.Response')
    def test_use_response_content(self, resp):
        parser_return_value = [FakeParser(), FakeParser()]

        transport = RequestsXmlrpcTransport('http://localhost')
        transport.getparser = MagicMock(return_value=parser_return_value)

        p = PropertyMock()
        type(resp).content = p
        transport.parse_response(resp)

        assert p.called, 'Response.content should be used instead'



from __future__ import print_function

import sys
import time

from xml.etree import ElementTree

from fake_virt import FakeVirt, FakeHandler


class EsxHandler(FakeHandler):
    def do_GET(self):
        print('[FakeEsx] GET', self.path)

    def do_POST(self):
        print("[FakeEsx] POST", self.path)
        if self.path == '/sdk':
            try:
                length = int(self.headers.getheader('content-length'))
            except AttributeError:
                length = int(self.headers.get('content-length'))
            data = self.rfile.read(length)
            xml = ElementTree.fromstring(data)
            body = xml.find('{http://schemas.xmlsoap.org/soap/envelope/}Body')
            root = body[0]
            print("[FakeEsx] post ", self.path, root.tag, self.server._data_version.value)

            if 'RetrieveServiceContent' in root.tag:
                self.write_file('esx', 'esx_retrieveservicecontent.xml')
            elif 'Login' in root.tag:
                if root.find('{urn:vim25}userName').text != 'A!bc3#\'"' or root.find('{urn:vim25}password').text != 'A!bc3#\'"':
                    self.send_error(401, 'Cannot complete login due to an incorrect user name or password.')
                self.write_file('esx', 'esx_loginresponse.xml')
            elif 'Logout' in root.tag:
                self.write_file('esx', 'esx_logoutresponse.xml')
            elif 'CreateFilter' in root.tag:
                self.write_file('esx', 'esx_createfilterresponse.xml')
            elif 'WaitForUpdatesEx' in root.tag:
                time.sleep(1)
                version = self.server._data_version.value
                self.write_file('esx', 'esx_waitforupdatesexresponse_%d.xml' % version)
            elif 'CancelWaitForUpdatesEx' in root.tag:
                self.write_file('esx', 'esx_cancelwaitforupdatesexresponse.xml')
            elif 'DestroyPropertyFilter' in root.tag:
                self.write_file('esx', 'esx_destroypropertyfilterresponse.xml')


class FakeEsx(FakeVirt):
    def __init__(self, port=None):
        super(FakeEsx, self).__init__(EsxHandler, port=port)
        self.server._data_version = self._data_version

if __name__ == '__main__':
    if len(sys.argv) >= 2:
        port = int(sys.argv[1])
    else:
        port = None

    esx = FakeEsx(port)
    esx.run()



# -*- coding: utf-8 -*-
from __future__ import print_function


"""
Module for abstraction of all virtualization backends, part of virt-who

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import time
from virtwho import log
from operator import itemgetter
from datetime import datetime
from threading import Thread, Event
import json
import hashlib
import re
import fnmatch
import six
from virtwho.config import NotSetSentinel, Satellite5DestinationInfo, \
    Satellite6DestinationInfo, DefaultDestinationInfo, VW_GLOBAL
from virtwho.manager import ManagerError, ManagerThrottleError, ManagerFatalError
from virtwho import MinimumSendInterval, MinimumJobPollInterval

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6 doesn't have OrderedDict, we need to have our own
    from virtwho.util import OrderedDict


class VirtError(Exception):
    pass


class Guest(object):
    """
    This class represents one virtualization guest running on some
    host/hypervisor.
    """

    STATE_UNKNOWN = 0      # unknown state
    STATE_RUNNING = 1      # running
    STATE_BLOCKED = 2      # blocked on resource
    STATE_PAUSED = 3       # paused by user
    STATE_SHUTINGDOWN = 4  # guest is being shut down
    STATE_SHUTOFF = 5      # shut off
    STATE_CRASHED = 6      # crashed
    STATE_PMSUSPENDED = 7  # suspended by guest power management

    def __init__(self,
                 uuid,
                 virt_type,
                 state,
                 hypervisorType=None):
        """
        Create new guest instance that will be sent to the subscription manager.

        `uuid` is unique identification of the guest.

        `virt` is a `Virt` class instance that represents virtualization hypervisor
        that owns the guest.

        `state` is a number that represents the state of the guest (stopped, running, ...)
        """
        self.uuid = uuid
        self.virtWhoType = virt_type
        self.state = state

    def __repr__(self):
        return 'Guest({0.uuid!r}, {0.virtWhoType!r}, {0.state!r})'.format(self)

    def toDict(self):
        d = OrderedDict((
            ('guestId', self.uuid),
            ('state', self.state),
            ('attributes', {
                'virtWhoType': self.virtWhoType,
                'active': 1 if self.state in (self.STATE_RUNNING, self.STATE_PAUSED) else 0
            }),
        ))
        return d


class Hypervisor(object):
    """
    A model for information about a hypervisor
    """

    CPU_SOCKET_FACT = 'cpu.cpu_socket(s)'
    HYPERVISOR_TYPE_FACT = 'hypervisor.type'
    HYPERVISOR_VERSION_FACT = 'hypervisor.version'
    HYPERVISOR_CLUSTER = 'hypervisor.cluster'
    SYSTEM_UUID_FACT = 'dmi.system.uuid'

    def __init__(self, hypervisorId, guestIds=None, name=None, facts=None):
        """
        Create a new Hypervisor that will be sent to subscription manager

        'hypervisorId': the unique identifier for this hypervisor

        'guestIds': a list of Guests

        'name': the hostname, if available
        """
        self.hypervisorId = hypervisorId
        self.guestIds = guestIds or []
        self.name = name
        self.facts = facts

    def __repr__(self):
        return 'Hypervisor({0.hypervisorId!r}, {0.guestIds!r}, {0.name!r}, {0.facts!r})'.format(self)

    def toDict(self):
        d = OrderedDict((
            ('hypervisorId', {'hypervisorId': self.hypervisorId}),
            ('name', self.name),
            ('guestIds', sorted([g.toDict() for g in self.guestIds], key=itemgetter('guestId')))
        ))
        if self.name is None:
            del d['name']
        if self.facts is not None:
            d['facts'] = self.facts
        return d

    def __str__(self):
        return str(self.toDict())

    def getHash(self):
        sortedRepresentation = json.dumps(self.toDict(), sort_keys=True)
        return hashlib.sha256(sortedRepresentation).hexdigest()


class AbstractVirtReport(object):
    """
    An abstract report from virt backend.
    """
    # The report was just collected, but is not yet being reported
    STATE_CREATED = 1
    # The report is being processed by server
    STATE_PROCESSING = 2
    # The report has been processed by server
    STATE_FINISHED = 3
    # Failed to process the report by server
    STATE_FAILED = 4
    # Processing the report on server was canceled
    STATE_CANCELED = 5

    def __init__(self, config, state=STATE_CREATED):
        self._config = config
        self._state = state

    def __repr__(self):
        return '{1}({0.config!r}, {0.state!r})'.format(self, self.__class__.__name__)

    @property
    def config(self):
        return self._config

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def hash(self):
        return hash(self)


class ErrorReport(AbstractVirtReport):
    """
    Report that virt backend fails. Used in oneshot mode to inform
    main thread that no data are coming.
    """


class DomainListReport(AbstractVirtReport):
    """
    Report from virt backend about list of virtual guests on given system.
    """
    def __init__(self, config, guests, hypervisor_id=None, state=AbstractVirtReport.STATE_CREATED):
        super(DomainListReport, self).__init__(config, state)
        self._guests = guests
        self._hypervisor_id = hypervisor_id

    def __repr__(self):
        return 'DomainListReport({0.config!r}, {0.guests!r}, {0.hypervisor_id!r}, {0.state!r})'.format(self)

    @property
    def guests(self):
        return self._guests

    @property
    def hypervisor_id(self):
        return self._hypervisor_id

    @property
    def hash(self):
        current_hash = json.dumps(
                sorted([g.toDict() for g in self.guests], key=itemgetter('guestId')),
                sort_keys=True)
        current_hash += str(self.hypervisor_id)
        return hashlib.sha256(current_hash.encode('utf-8')).hexdigest()


class HostGuestAssociationReport(AbstractVirtReport):
    """
    Report from virt backend about host/guest association on given hypervisor.
    """
    def __init__(self, config, assoc, state=AbstractVirtReport.STATE_CREATED,
                 exclude_hosts=None, filter_hosts=None):
        super(HostGuestAssociationReport, self).__init__(config, state)
        self._assoc = assoc
        if exclude_hosts is None:
            try:
                exclude_hosts = self._config['exclude_hosts']
            except KeyError:
                # We do not have a config that has this attribute
                pass
        if filter_hosts is None:
            try:
                filter_hosts = self._config['filter_hosts']
            except KeyError:
                # We do not have a config with this attribute
                pass
        self.exclude_hosts = exclude_hosts
        self.filter_hosts = filter_hosts
        try:
            self.filter_type = self._config['filter_type']
        except KeyError:
            # FIXME: default value should be there
            self.filter_type = None

    def __repr__(self):
        return 'HostGuestAssociationReport({0.config!r}, {0._assoc!r}, {0.state!r})'.format(self)

    def _filter(self, host, filterlist):
        for i in filterlist:
            if self.filter_type is None:
                if fnmatch.fnmatch(host.lower(), i.lower()):
                    # match is found
                    return True
                try:
                    if re.match("^" + i + "$", host, re.IGNORECASE):
                        # match is found
                        return True
                except:
                    pass
            elif self.filter_type == "wildcards":
                if fnmatch.fnmatch(host.lower(), i.lower()):
                    # match is found
                    return True
            elif self.filter_type == "regex":
                try:
                    if re.match("^" + i + "$", host, re.IGNORECASE):
                        # match is found
                        return True
                except:
                    pass
        # no match
        return False

    @property
    def association(self):
        # Apply filter
        logger = log.getLogger(name='virt', queue=False)
        assoc = []
        for host in self._assoc['hypervisors']:
            if self.exclude_hosts is not None and self.exclude_hosts != NotSetSentinel:
                if self._filter(host.hypervisorId, self.exclude_hosts):
                    logger.debug("Skipping host '%s' because its ID was excluded by filter '%s'" %
                                 (host.hypervisorId, self.exclude_hosts))
                    continue
                else:
                    logger.debug("Host %s passed filter %s" % (host.hypervisorId, self.exclude_hosts))

            if self.filter_hosts is not None and self.filter_hosts != NotSetSentinel:
                if self._filter(host.hypervisorId, self.filter_hosts):
                    logger.debug("Host %s passed filter %s" % (host.hypervisorId, self.filter_hosts))
                else:
                    logger.debug("Skipping host '%s' because its ID was not included in filter '%s'" %
                                 (host.hypervisorId, self.filter_hosts))
                    continue

            assoc.append(host)
        return {'hypervisors': assoc}

    @property
    def serializedAssociation(self):
        id_getter = itemgetter('hypervisorId')
        return {
            'hypervisors': sorted([h.toDict() for h in self.association['hypervisors']],
                                  key=lambda x: id_getter(id_getter(x)))
        }

    @property
    def hash(self):
        return hashlib.sha256(json.dumps(self.serializedAssociation, sort_keys=True).encode('utf-8')).hexdigest()


class IntervalThread(Thread):
    def __init__(self, logger, config, source=None, dest=None,
                 terminate_event=None, interval=None, oneshot=False):
        self.logger = logger
        self.config = config
        self.source = source
        self.dest = dest
        self._internal_terminate_event = Event()
        self.terminate_event = terminate_event or self._internal_terminate_event
        self.interval = interval
        self._oneshot = oneshot
        self.delta_time = 1.0
        super(IntervalThread, self).__init__()

    def wait(self, wait_time):
        """
        Wait `wait_time` seconds, could be interrupted by setting _terminate_event or _internal_terminate_event.
        """
        total_time = 0.0
        while total_time < wait_time:
            if self.is_terminated():
                break
            time.sleep(self.delta_time)
            total_time += self.delta_time

    def is_terminated(self):
        """

        @return: Returns true if either the internal terminate event is set or
                 the terminate event given in the init is set
        """
        return self._internal_terminate_event.is_set() or \
            self.terminate_event.is_set()

    def stop(self):
        """
        Causes this thread to stop at the next idle moment
        """
        self._internal_terminate_event.set()

    def _run(self):
        """
        This method could be reimplemented in subclass to provide
        it's own way of waiting for changes (like event monitoring)
        """
        self.prepare()
        while not self.is_terminated():
            start_time = datetime.now()
            data_to_send = self._get_data()
            self._send_data(data_to_send)
            if self._oneshot:
                self._internal_terminate_event.set()
                break
            end_time = datetime.now()

            delta = end_time - start_time
            # for python2.6, 2.7 has total_seconds method
            delta_seconds = ((
                             delta.days * 86400 + delta.seconds) * 10 ** 6 +
                             delta.microseconds) / 10 ** 6

            wait_time = self.interval - int(delta_seconds)

            if wait_time < 0:
                self.logger.debug(
                    "Getting the data took longer than the configured "
                    "interval. Trying again immediately.")
                continue

            self.wait(wait_time)

    def _get_data(self):
        """
        This method gathers data from the source provided to the thread
        @return: The data from the source
        """
        raise NotImplementedError("Should be implemented in subclasses")

    def _send_data(self, data_to_send):
        """
        @param data_to_send: The data to be given to the dest
        """
        raise NotImplementedError("Should be implemented in subclasses")

    def run(self):
        '''
        Wrapper around `_run` method that just catches the error messages.
        '''
        self.logger.debug("Thread '%s' started", self.config.name)
        try:
            while not self.is_terminated():
                has_error = False
                try:
                    self._run()
                except VirtError as e:
                    if not self.is_terminated():
                        self.logger.error("Thread '%s' fails with error: %s",
                                          self.config.name, str(e))
                        has_error = True
                except Exception:
                    if not self.is_terminated():
                        self.logger.exception("Thread '%s' fails with "
                                              "exception:", self.config.name)
                        has_error = True

                if self.is_terminated():
                    self.logger.debug("Thread '%s' terminated",
                                      self.config.name)
                    self._internal_terminate_event.set()
                    return

                if has_error:
                    self._send_data(ErrorReport(self.config))

                if self._oneshot:
                    self.logger.debug("Thread '%s' stopped after running once",
                                      self.config.name)
                    self._internal_terminate_event.set()
                    return

                self.logger.info("Waiting %s seconds before performing action"
                                 " again '%s'", self.interval, self.config.name)
                self.wait(self.interval)
        except KeyboardInterrupt:
            self.logger.debug("Thread '%s' interrupted", self.config.name)
            self.cleanup()

    def cleanup(self):
        '''
        Perform cleaning up actions before termination.
        '''
        pass

    def prepare(self):
        """
        Do pre-mainloop initialization of the source and dest,
        for example logging in.
        """
        pass

    @staticmethod
    def handle_429(retry_after, number_of_failures):
        """
        @param retry_after: The value of the Retry-After header (if included)
        @type retry_after: int
        @param number_of_failures: The number of failures that have happened attempting the same
        request.
        @type retry_after: int

        @return: The number of seconds that should be waited before retrying
        @rtype: int
        """
        if retry_after is not None:
            try:
                return int(retry_after)
            except (TypeError, ValueError):
                # if the retry_after value is not convertable to an int, don't use it
                pass
        # If there is no good retry-after value to use, use twice the polling interval
        return MinimumJobPollInterval * 2


class DestinationThread(IntervalThread):
    """
    This class is a thread that pulls reports from the datastore and sends them
    to the actual destination (candlepin, Satellite, etc) using a manager
    object.

    This class should work so long as the destination is a Manager object.
    """

    def __init__(self, logger, config, source_keys=None, options=None,
                 source=None, dest=None, terminate_event=None, interval=None,
                 oneshot=False):
        """
        @param source_keys: A list of keys to be used to retrieve info from
        the source
        @type source_keys: list

        @param source: The source to pull from
        @type source: Datastore

        @param dest: The destination object to use to actually send the data
        @type dest: Manager
        """
        if not isinstance(source_keys, list):
            raise ValueError("Source keys must be a list")
        self.is_initial_run = True
        self.source_keys = source_keys
        self.last_report_for_source = {}  # Source_key to hash of last report
        self.submitted_report_and_hash_for_source = {}  # Source key to submitted batch report and hash
        self.options = options
        self.reports_to_print = []  # A list of reports we would send but are
        #  going to print instead, to be used by the owner of the thread
        # after the thread has been killed
        super(DestinationThread, self).__init__(logger, config, source=source,
                                                dest=dest,
                                                terminate_event=terminate_event,
                                                interval=interval,
                                                oneshot=oneshot)
        # This is used when there is some reason to modify how long we wait
        # EX when we get a 429 back from the server, this value will be the
        # value of the retry_after header.
        self.interval_modifier = 0

    def _get_data(self):
        """
        Gets the latest report from the source for each source_key
        @return: dict
        """
        if self.is_initial_run:
            return self._get_data_initial()
        return self._get_data_common(self.source_keys)

    def _get_data_common(self, source_keys, ignore_duplicates=True, log_missing_reports=True):
        reports = {}
        for source_key in source_keys:
            report = self.source.get(source_key, NotSetSentinel)
            if report is None or report is NotSetSentinel:
                if log_missing_reports:
                    self.logger.debug("No report available for source: %s" %
                                      source_key)
                continue
            if source_key in self.submitted_report_and_hash_for_source:
                submitted_report = self.submitted_report_and_hash_for_source[source_key][0]
                submitted_hash = self.submitted_report_and_hash_for_source[source_key][1]
                self.check_report_status(submitted_report)
                self.submitted_report_and_hash_for_source.pop(source_key)
                if submitted_report.state == AbstractVirtReport.STATE_FINISHED:
                    self.last_report_for_source[source_key] = submitted_hash
                    if ignore_duplicates and report.hash == submitted_hash:
                        self.logger.debug('Duplicate report found for config "%s", ignoring',
                                          report.config.name)
                        continue
                elif submitted_report.state == AbstractVirtReport.STATE_PROCESSING:
                    # still processing, skip this fabric on this cycle
                    self.logger.warning('Job %s has not finished processing. Will check after next interval.',
                                        str(submitted_report.job_id))
                    self.submitted_report_and_hash_for_source[source_key] = (submitted_report, submitted_hash)
                    continue
            elif ignore_duplicates and report.hash == self.last_report_for_source.get(source_key,
                                                                                    None):
                self.logger.debug('Duplicate report found for config "%s", ignoring',
                                  report.config.name)
                continue
            reports[source_key] = report
        return reports

    def _get_data_initial(self):
        """
        Waits for each source in self.source_keys to have a value returned from the datastore. This
        does not check to see if the data has been previously sent. This method will wait for a
        maximum of the self.interval period of time, then return whatever it has gathered thus far.
        If data for each source key is gathered before the interval has expired, this method will
        return.
        @return: dict
        """
        reports = {}
        while not reports and not self.is_terminated():
            source_keys_remaining = set(self.source_keys)
            time_waited = 0
            while len(source_keys_remaining) > 0 and time_waited < self.interval and not self.is_terminated():
                found_reports = self._get_data_common(source_keys_remaining,
                                                      ignore_duplicates=False,
                                                      log_missing_reports=False)
                reports.update(found_reports)
                source_keys_remaining.difference_update(found_reports.keys())
                if len(source_keys_remaining) > 0:
                    sleep_time = 1
                    time.sleep(sleep_time)
                    time_waited += sleep_time
        self.is_initial_run = False
        return reports

    def _send_data(self, data_to_send):
        """
        Processes the data_to_send and sends it using the dest object.
        @param data_to_send: A dict of source_keys, report
        @type: dict
        """

        if not isinstance(data_to_send, ErrorReport):
            # Ping all hypervisors for this reporter to update check in time regardless of change
            self.dest.hypervisorHeartbeat(config=self.config, options=self.options)

        if not data_to_send:
            self.logger.debug('No data to send, waiting for next interval')
            return
        if isinstance(data_to_send, ErrorReport):
            self.logger.info('Error report received')
            if data_to_send.config.name != self.config.name:
                self.stop()
            return

        all_hypervisors = []  # All the Host-guest mappings together
        domain_list_reports = []  # Source_keys of DomainListReports
        reports_batched = []  # Source_keys of reports to be sent as one
        sources_sent = []  # Sources we have dealt with this run
        sources_erred = []  # Sources with some problems

        total_hypervisors = 0
        total_guests = 0

        # Reports of different types are handled differently
        for source_key, report in data_to_send.items():
            if getattr(self.config, 'owner', None) is None and \
                    isinstance(report, HostGuestAssociationReport):
                # If the owner on our config is not defined, set it to the first report that
                # we've found. This should be ok because destination threads should not be run for
                # more than one owner.
                self.config['owner'] = report.config['owner']
            if isinstance(report, DomainListReport):
                # These are sent one at a time to the destination
                domain_list_reports.append(source_key)
                continue
            if isinstance(report, HostGuestAssociationReport):
                # These reports are put into one report to send at once
                all_hypervisors.extend(report.association['hypervisors'])
                # Keep track of those reports that we have
                reports_batched.append(source_key)
                # Print information about host-to-quest mapping for this report
                mapping = report.association
                hypervisor_count = len(mapping['hypervisors'])
                guest_count = sum(len(hypervisor.guestIds) for hypervisor in mapping['hypervisors'])
                self.logger.info('Hosts-to-guests mapping for config "%s": %d hypervisors and %d guests found',
                                 report.config.name, hypervisor_count, guest_count)
                total_hypervisors += hypervisor_count
                total_guests += guest_count
                continue
            if isinstance(report, ErrorReport):
                # These indicate an error that came from this source
                # Log it and move along.
                # if it was recoverable we'll get something else next time.
                # if it was not recoverable we'll see this again from this
                # source. Thus we'll just log this at the debug level.
                self.logger.debug('ErrorReport received for source: %s' % source_key)
                if self._oneshot:
                    # Consider this source dealt with if we are in oneshot mode
                    sources_erred.append(source_key)

        # Modify the batched dict to be in the form expected for
        # HostGuestAssociationReports
        all_hypervisors_dict = {'hypervisors': all_hypervisors}
        batch_host_guest_report = HostGuestAssociationReport(self.config, all_hypervisors_dict)

        if all_hypervisors:
            result = None
            # Try to actually do the checkin whilst being mindful of the
            # rate limit (retrying where necessary)
            num_429_received = 0
            while result is None and not self.is_terminated():
                try:
                    self.logger.info('Sending updated Host-to-guest mapping to "{owner}" including '
                                     '{num_hypervisors} hypervisors and {num_guests} '
                                     'guests'.format(owner=self.config['owner'],
                                                     num_hypervisors=total_hypervisors,
                                                     num_guests=total_guests))
                    result = self.dest.hypervisorCheckIn(
                            batch_host_guest_report,
                            options=self.options)
                    break
                except ManagerThrottleError as e:
                    if self._oneshot:
                        self.logger.debug('429 encountered while performing hypervisor checkin in '
                                          'oneshot mode, not retrying')
                        sources_erred.extend(reports_batched)
                        break
                    num_429_received += 1
                    retry_after = self.handle_429(e.retry_after, num_429_received)
                    self.logger.debug("429 encountered while performing "
                                      "hypervisor check in.\n"
                                      "Trying again in "
                                      "%s", retry_after)
                    self.interval_modifier = retry_after
                except (ManagerError, ManagerFatalError) as err:
                    self.logger.exception("Error during hypervisor "
                                          "checkin: %s" % err)
                    if self._oneshot:
                        sources_erred.extend(reports_batched)
                    break
                self.wait(wait_time=self.interval_modifier)
                self.interval_modifier = 0

            if result:
                for source_key in reports_batched:
                    self.submitted_report_and_hash_for_source[source_key] =\
                        (batch_host_guest_report, data_to_send[source_key].hash)

        # Send each Domain Guest List Report if necessary
        for source_key in domain_list_reports:
            report = data_to_send[source_key]
            if not self.options[VW_GLOBAL]['print']:
                retry = True
                num_429_received = 0
                while retry and not self.is_terminated():  # Retry if we encounter a 429
                    try:
                        self.dest.sendVirtGuests(report, options=self.options)
                        sources_sent.append(source_key)
                        self.last_report_for_source[source_key] = data_to_send[
                            source_key].hash
                        retry = False
                    except ManagerThrottleError as e:
                        if self._oneshot:
                            self.logger.info(
                                'Rate limit exceeded while sending host-guest mapping, please see: '
                                'https://access.redhat.com/solutions/2212941. '
                                'Running in one-shot mode, not retrying.')
                            sources_erred.append(source_key)
                            break
                        num_429_received += 1
                        retry_after = self.handle_429(e.retry_after, num_429_received)
                        self.logger.info(
                            'Rate limit exceeded while sending host-guest mapping, please see: '
                            'https://access.redhat.com/solutions/2212941. '
                            'Retrying after: %s seconds.', retry_after)
                        self.wait(wait_time=retry_after)
                    except (ManagerError, ManagerFatalError):
                        self.logger.exception("Fatal error during send virt "
                                              "guests: ")
                        if self._oneshot:
                            sources_erred.append(source_key)
                        retry = False  # Only retry on 429

        # Were all sources handled at lease by one report?
        all_sources_handled = all((source_key in sources_sent or source_key in sources_erred)
                                  for source_key in self.source_keys)

        # Terminate this thread if we have sent one report for each source
        if all_sources_handled and self._oneshot:
                if not self.options[VW_GLOBAL]['print']:
                    self.logger.debug('At least one report for each connected source has been sent. Terminating.')
                else:
                    self.logger.debug('All info to print has been gathered. Terminating.')
                self.stop()

        if self._oneshot:
            # Remove sources we have sent (or dealt with) so that we don't
            # do extra work on the next run, should we have missed any sources
            self.source_keys = [source_key for source_key in self.source_keys
                                if source_key not in sources_sent]
        return

    def check_report_status(self, report):
        """
        Checks at the server for the state of the previously submitted job. The state is recorded
        in the passed-in report
        """
        self.logger.debug("Existing report state: %s" % report.state)
        num_429_received = 0
        first_attempt = True
        while not report.state or report.state == AbstractVirtReport.STATE_CREATED\
            or report.state == AbstractVirtReport.STATE_PROCESSING and first_attempt:
            if self.interval_modifier != 0:
                wait_time = self.interval_modifier
                self.interval_modifier = 0
            elif not first_attempt:
                wait_time = MinimumJobPollInterval * 2
            else:
                wait_time = MinimumJobPollInterval

            self.wait(wait_time=wait_time)

            try:
                self.dest.check_report_state(report)
            except ManagerThrottleError as e:
                if self._oneshot:
                    self.logger.debug('429 encountered when checking job state in '
                                      'oneshot mode, not retrying')
                    break
                retry_after = self.handle_429(e.retry_after, num_429_received)
                self.logger.debug('429 encountered while checking job '
                                  'state, checking again in "%s"', retry_after)
                self.interval_modifier = retry_after
            except (ManagerError, ManagerFatalError):
                self.logger.exception("Error during job check: ")
                break
            # If we get here and have to try again, it's not our first rodeo...
            first_attempt = False


class Satellite5DestinationThread(DestinationThread):

    def _send_data(self, data_to_send):
        """
        Processes the data_to_send and sends it using the dest object.
        @param data_to_send: A dict of source_keys, report
        @type: dict
        """
        if not data_to_send:
            self.logger.debug('No data to send, waiting for next interval')
            return
        if isinstance(data_to_send, ErrorReport):
            self.logger.info('Error report received, shutting down')
            self.stop()
            return
        sources_sent = []  # Sources we have dealt with this run
        sources_erred = []  # Sources that have had some error this run

        # Reports of different types are handled differently
        for source_key, report in data_to_send.items():
            if isinstance(report, DomainListReport):
                self.logger.warning("virt-who does not support sending local"
                                    "hypervisor data to satellite; use "
                                    "rhn-virtualization-host "
                                    "instead; Dropping offending source: '%s'",
                                    source_key)
                # Do not attempt to send such reports, satellite 5 does not
                # know what to do with such reports. Since we do not know
                # what to do, and virt backends will not change their output
                # without a restart, drop this source from those that we check.
                sources_erred.append(source_key)
                continue
            if isinstance(report, HostGuestAssociationReport):
                # We cannot (effectively) batch reports to be checked in in
                # one communication via Satellite 5. As such we'll just do a
                # hypervisor check in for each report of that type.
                result = None
                num_429_received = 0
                while result is None and not self.is_terminated():
                    try:
                        result = self.dest.hypervisorCheckIn(
                                report,
                                options=self.options)
                        self.last_report_for_source[source_key] = report.hash
                        sources_sent.append(source_key)
                        break
                    except ManagerThrottleError as e:
                        if self._oneshot:
                            self.logger.debug('429 encountered during hypervisor checkin in '
                                             'oneshot mode, not retrying')
                            sources_erred.append(source_key)
                            break
                        num_429_received += 1
                        retry_after = self.handle_429(e.retry_after, num_429_received)
                        self.logger.debug("429 encountered while performing "
                                          "hypervisor check in.\n"
                                          "Trying again in "
                                          "%s", retry_after)
                        self.interval_modifier = retry_after
                    except ManagerFatalError:
                        self.logger.exception("Fatal error during hypervisor "
                                              "checkin: ")
                        sources_erred.append(source_key)
                        break
            if isinstance(report, ErrorReport):
                # These indicate an error that came from this source
                # Log it and move along.
                # if it was recoverable we'll get something else next time.
                # if it was not recoverable we'll see this again from this
                # source. Thus we'll just log this at the debug level.
                self.logger.debug('ErrorReport received for source: %s' % source_key)
                if self._oneshot:
                    # Consider this source dealt with if we are in oneshot mode
                    sources_sent.append(source_key)

        # Terminate this thread if we have sent one report for each source
        if all(source_key in sources_sent for source_key in self.source_keys)\
                and self._oneshot:
            if not self.options[VW_GLOBAL]['print']:
                self.logger.debug('At least one report for each connected '
                                  'source has been sent. Terminating.')
            else:
                self.logger.debug('All info to print has been gathered. '
                                  'Terminating.')
            self.stop()
        if self._oneshot:
            # Remove sources we have sent (or dealt with) so that we don't
            # do extra work on the next run, should we have missed any sources
            self.source_keys = [source_key for source_key in self.source_keys
                                if source_key not in sources_sent]
        return


class Virt(IntervalThread):
    """
    Virtualization backend abstract class.

    This class must be inherited for each of the virtualization backends.

    Run `start` method to start obtaining data about virtual guests. The data
    will be pushed to the dest(ination) that is parameter of the `__init__`
    method.
    """

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False):
        super(Virt, self).__init__(logger, config, dest=dest,
                                   terminate_event=terminate_event,
                                   interval=interval, oneshot=oneshot)

    @classmethod
    def __subclasses_list(cls):
        # Imports can't be top-level, it would be circular dependency
        import virtwho.virt.libvirtd  # flake8: noqa
        import virtwho.virt.esx  # flake8: noqa
        import virtwho.virt.xen  # flake8: noqa
        import virtwho.virt.rhevm  # flake8: noqa
        import virtwho.virt.vdsm  # flake8: noqa
        import virtwho.virt.hyperv  # flake8: noqa
        import virtwho.virt.fakevirt  # flake8: noqa
        import virtwho.virt.kubevirt  # flake8: noqa
        import virtwho.virt.ahv       # flake8: noqa

        return [subcls for subcls in cls.__subclasses__()]

    @classmethod
    def from_config(cls, logger, config, dest,
                    terminate_event=None, interval=None, oneshot=False):
        """
        Create instance of inherited class based on the config.
        """

        for subcls in cls.__subclasses_list():
            if config['type'] == subcls.CONFIG_TYPE:
                return subcls(logger, config, dest,
                              terminate_event=terminate_event,
                              interval=interval, oneshot=oneshot)
        raise KeyError("Invalid config type: %s" % config['type'])

    @classmethod
    def hypervisor_types(cls):
        """
        Return list of textual representations of supported hypervisors
        """

        return [subcls.CONFIG_TYPE for subcls in cls.__subclasses_list() if subcls.CONFIG_TYPE != 'fake']

    def start_sync(self):
        '''
        This method is same as `start()` but runs synchronously, it does NOT
        create new thread.

        Use it only in specific cases!
        '''
        self._run()

    def _get_report(self):
        if self.isHypervisor():
            return HostGuestAssociationReport(self.config, self.getHostGuestMapping())
        else:
            return DomainListReport(self.config, self.listDomains())

    # TODO: Reimplement each virt subclass as a source
    def _get_data(self):
        """
        Gathers the data from the source.
        Could be overridden to specify how to get data other than a report.
        For example in destination threads.
        @return: The data from the source to be passed along to the dest
        """
        return self._get_report()

    def _send_data(self, data_to_send):
        if self.is_terminated():
            return
        self.logger.info('Report for config "%s" gathered, placing in '
                          'datastore', data_to_send.config.name)
        self.dest.put(self.config.name, data_to_send)

    def isHypervisor(self):
        """
        Return True if the virt instance represents hypervisor environment
        otherwise it represents just one virtual server.
        """
        return True

    def getHostGuestMapping(self):
        '''
        If subclass doesn't reimplement the `_run` method, it should
        reimplement either this method or `listDomains` method, based on
        return value of isHypervisor method.
        '''
        raise NotImplementedError('This should be reimplemented in subclass')

    def listDomains(self):
        '''
        If subclass doesn't reimplement the `_run` method, it should
        reimplement either this method or `getHostGuestMapping` method, based on
        return value of isHypervisor method.
        '''
        raise NotImplementedError('This should be reimplemented in subclass')

    @staticmethod
    def _to_unicode(value):
        try:
            return six.text_type(value, 'utf-8')
        except TypeError:
            return value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = self._to_unicode(value)

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = self._to_unicode(value)


info_to_destination_class = {
    Satellite5DestinationInfo: Satellite5DestinationThread,
    Satellite6DestinationInfo: DestinationThread,
    DefaultDestinationInfo: DestinationThread,
}



from __future__ import print_function

import time

from fake_virt import FakeVirt, FakeHandler


class Rhevm4Handler(FakeHandler):
    def do_GET(self):
        time.sleep(0.1)
        print(("DO GET", self.path))

        if self.path == '/ovirt-engine/api':
            self.write_file('rhevm', 'rhev4_api.xml')
        elif self.path == '/ovirt-engine/api/clusters':
            self.check_version_header()
            self.write_file('rhevm', 'rhevm_clusters.xml')
        elif self.path == '/ovirt-engine/api/hosts':
            self.check_version_header()
            self.write_file('rhevm', 'rhevm_hosts.xml')
        elif self.path == '/ovirt-engine/api/vms':
            self.check_version_header()
            self.write_file('rhevm', 'rhevm_vms_%d.xml' % self.server._data_version.value)
        else:
            self.send_response(404)
            self.end_headers()

    def check_version_header(self):
        if not self.headers['Version'] == '3':
            self.send_response(400, 'Version header mismatch')


class FakeRhevm4(FakeVirt):
    def __init__(self, port=None):
        super(FakeRhevm4, self).__init__(Rhevm4Handler, port=port)
        self.server._data_version = self._data_version

if __name__ == '__main__':
    rhevm = FakeRhevm4(port=8443)
    rhevm.run()



# -*- coding: utf-8 -*-
from __future__ import print_function

#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""
Test validating of GlobalSection.
"""

from base import TestBase

from virtwho.config import GlobalSection, str_to_bool, MinimumSendInterval, DefaultInterval
from virtwho.log import DEFAULT_LOG_DIR

# Values used for testing GlobalConfigSection
GLOBAL_SECTION_VALUES = {
    'debug': True,
    'oneshot': 'true',
    'print': 'false',
    'wrong_bool_value': 'nein',
    'configs': ['local_libvirt.conf'],
    'log_file': 'my_custom.log',
    'interval': '120'
}


class TestGlobalConfigSection(TestBase):
    """
    Test base for testing class GlobalSection
    """

    def setUp(self):
        self.init_global_config_section()

    def init_global_config_section(self):
        """
        Method executed before each unit test
        """
        self.global_config = GlobalSection('global', None)
        # We need to set values using this way, because we need
        # to trigger __setitem__ of global_config
        for key, value in GLOBAL_SECTION_VALUES.items():
            self.global_config[key] = value

    def test_validate_debug_value_bool(self):
        """
        Test validating of correct bool value 
        """
        result = self.global_config._validate_str_to_bool('debug')
        self.assertIsNone(result)
        value = self.global_config.get('debug')
        self.assertIs(value, True)

    def test_validate_oneshot_value_string_true(self):
        """
        Test validating of correct bool value stored as string ('true') 
        """
        result = self.global_config._validate_str_to_bool('oneshot')
        self.assertIsNone(result)
        value = self.global_config.get('oneshot')
        self.assertIs(value, True)

    def test_validate_print_value_string_false(self):
        """
        Test validating of correct bool value stored as string ('false') 
        """
        result = self.global_config._validate_str_to_bool('print')
        self.assertIsNone(result)
        value = self.global_config.get('print')
        self.assertIs(value, False)

    def test_validate_wrong_bool_value(self):
        """
        Test validating of wrong string representing bool ('nein') 
        """
        self.assertRaises(ValueError, str_to_bool, self.global_config['wrong_bool_value'])
        result = self.global_config._validate_str_to_bool('wrong_bool_value')
        self.assertIsNotNone(result)
        self.assertIsInstance(result, tuple)

    def test_validate_non_existing_key(self):
        """
        Test validation of non-existing key 
        """
        result = self.global_config._validate_str_to_bool('does_not_exist')
        self.assertIsNotNone(result)
        self.assertIsInstance(result, tuple)

    def test_validate_string(self):
        """
        Test validation of correct string value 
        """
        result = self.global_config._validate_non_empty_string('log_file')
        self.assertIsNone(result)

    def test_validate_interval(self):
        """
        Test validation of time interval
        """
        result = self.global_config._validate_interval('interval')
        self.assertIsNone(result)

    def test_validate_wrong_interval(self):
        """
        Test validation of wrong time interval
        """
        self.global_config['interval'] = '10'
        self.global_config.validate()
        interval = self.global_config.get('interval')
        # The behavior of this has changed. We now replace any invalid value with the default
        # No special cases
        self.assertIs(interval, DefaultInterval)

    def test_validate_missing_interval(self):
        """
        Test validation of wrong time interval
        """
        self.global_config = GlobalSection('global', None)
        for key, value in GLOBAL_SECTION_VALUES.items():
            if key != 'interval':
                self.global_config[key] = value
        result = self.global_config._validate_interval('interval')
        self.assertIsNotNone(result)
        interval = self.global_config['interval']
        self.assertIs(interval, DefaultInterval)

    def test_validate_configs(self):
        """
        Test validation of configs (list of paths to config files)
        """
        result = self.global_config._validate_configs()
        self.assertIsNone(result)

    def test_validate_section_values(self):
        """
        Test validation of all config values 
        """
        validate_messages = self.global_config.validate()
        # TODO: use following expected messages, when format of warning/error messages will be settled down.
        # expected_results = [
        #     ('warning', 'Ignoring unknown configuration option "wrong_bool_value"'),
        #     (
        #         'warning',
        #         'log_per_config must be a valid boolean, using default. See man virt-who-config for more info'
        #     ),
        #     ('warning', 'Value for reporter_id not set, using default'),
        #     ('warning', 'Value for log_dir not set, using default'),
        # ]
        self.assertGreater(len(validate_messages), 0)
        # self.assertEqual(expected_results, validate_messages)

    def test_validate_default_value(self):
        """
        Test the validate method will set default value, when an option
        is missing and default value exist in DEFAULT
        """
        self.global_config.validate()
        default_value = self.global_config.get('log_dir')
        self.assertIs(default_value, DEFAULT_LOG_DIR)



#!/usr/bin/python2

from __future__ import print_function
import sys
import os
from setuptools import setup, find_packages
from distutils.command.install import install
from shutil import copy, copyfileobj
import gzip


def install_file(filename, destination, file_mode=0o644, dir_mode=0o755):
    install_dir = os.path.dirname(destination)
    if not os.path.isdir(install_dir):
        os.makedirs(install_dir, dir_mode)
    copy(filename, destination)
    os.chmod(destination, file_mode)


class InstallSystemd(install):
    description = 'install systemd service'

    def run(self, *args, **kwargs):
        root = self.root or ''
        if self.prefix != '/usr':
            print("systemd service has to be installed to /usr")
            sys.exit(1)
        install_file('virt-who.service', '{root}/usr/lib/systemd/system/virt-who.service'.format(root=root))


class InstallUpstart(install):
    description = 'install upstart service'

    def run(self, *args, **kwargs):
        root = self.root or ''
        install_file('virt-who-initscript', '{root}/etc/rc.d/init.d/virt-who'.format(root=root), file_mode=0o755)


class InstallManPages(install):
    description = 'install manual pages'

    MAN_PAGES = (
        ('virt-who', '8'),
        ('virt-who-config', '5'),
        ('virt-who-password', '8'),
    )

    def run(self, *args, **kwargs):
        root = self.root or ''
        # gzip embeds output filename and it breaks rpmbuild
        # we need to use relative name
        old_wd = os.getcwd()
        os.chdir(root)
        for name, number in self.MAN_PAGES:
            filename = '{old_wd}/{name}.{number}'.format(old_wd=old_wd, name=name, number=number)
            dirname = 'usr/share/man/man{number}'.format(number=number)
            if not os.path.isdir(dirname):
                os.makedirs(dirname, 0o755)
            outfile = '{dirname}/{name}.{number}.gz'.format(
                dirname=dirname, name=name, number=number)

            with open(filename, 'rb') as f_in:
                # gzip in py26 doesn't support contextmanager
                f_out = gzip.open(outfile, 'wb')
                try:
                    copyfileobj(f_in, f_out)
                finally:
                    f_out.close()
        os.chdir(old_wd)


class InstallConfig(install):
    description = 'install configuration files'

    FILES = (
        ('virt-who.conf', '{root}/etc/sysconfig/virt-who'),
        ('template.conf', '{root}/etc/virt-who.d/template.conf'),
        ('template-general.conf', '{root}/etc/virt-who.conf'),
    )

    def run(self, *args, **kwargs):
        root = self.root or ''
        for origname, output in self.FILES:
            install_file(origname, output.format(root=root))


version = {}
with open('virtwho/version.py') as ver_file:
    exec(ver_file.read(), version)

setup(
    name='virt-who',
    version='0.25.6',
    description='virt-who is agent for reporting virtual guest IDs to subscription manager.',
    # long_description=open('README.md').read(),
    author='Radek Novacek',
    author_email='rnovacek@redhat.com',
    license='LICENSE',
    url='https://github.com/virt-who/virt-who',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'virt-who = virtwho.__main__:main',
            'virt-who-password = virtwho.password.__main__:main'
        ]
    },
    include_package_data=True,
    package_data={
        'virtwho.virt.esx': ['vimServiceMinimal.wsdl'],
    },
    cmdclass={
        'install_systemd': InstallSystemd,
        'install_upstart': InstallUpstart,
        'install_man_pages': InstallManPages,
        'install_config': InstallConfig,
    },
)



# -*- coding: utf-8 -*-
from __future__ import print_function
"""
Module for accessing vdsm, part of virt-who

Parts of this file is based on rhn-virtualization from spacewalk
https://github.com/spacewalkproject/spacewalk/tree/master/client/tools/rhn-virtualization

Copyright (C) 2011 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import re
from six.moves import xmlrpc_client
from six.moves.configparser import SafeConfigParser, NoSectionError, NoOptionError
import os
import subprocess
import ssl


from virtwho.virt import Virt, Guest
from virtwho.config import VirtConfigSection
from virtwho.virt.vdsm.jsonrpc import JsonRpcClient


class VdsmError(Exception):
    pass


VDSM_STATE_TO_GUEST_STATE = {
    'Down': Guest.STATE_SHUTOFF,
    'Migration Destination': Guest.STATE_SHUTOFF,
    'Migration Source': Guest.STATE_SHUTINGDOWN,
    'Paused': Guest.STATE_PAUSED,
    'Powering down': Guest.STATE_SHUTINGDOWN,
    'RebootInProgress': Guest.STATE_SHUTOFF,
    'Restoring state': Guest.STATE_SHUTOFF,
    'Saving State': Guest.STATE_SHUTOFF,
    'Up': Guest.STATE_RUNNING,
    'WaitForLaunch': Guest.STATE_SHUTOFF,
    'Powering up': Guest.STATE_SHUTOFF
}

VDSM_JSONRPC_TIMEOUT = 300  # 5 minutes, borrowed from esx MAX_WAIT_TIME


class VdsmConfigSection(VirtConfigSection):
    """
    This class is used for validation of vdsm virtualization backend
    section(s). It tries to validate options and combination of options that
    are specific for this virtualization backend. In specific, it attempts to read
    the given file and produces error messages if it is not usable.
    """
    VIRT_TYPE = 'vdsm'

    def __init__(self, section_name, wrapper, *args, **kwargs):
        super(VdsmConfigSection, self).__init__(section_name, wrapper, *args, **kwargs)


class Vdsm(Virt):
    """
    Class for interacting with vdsmd daemon.
    """
    CONFIG_TYPE = "vdsm"

    def __init__(self, logger, config, dest, terminate_event=None,
                 interval=None, oneshot=False):
        super(Vdsm, self).__init__(logger, config, dest,
                                   terminate_event=terminate_event,
                                   interval=interval,
                                   oneshot=oneshot)
        self._read_config("/etc/vdsm/vdsm.conf")
        self.xmlrpc_client = None
        self.jsonrpc_client = None

    def isHypervisor(self):
        return False

    def _read_config(self, config_name):
        parser = SafeConfigParser()
        parser.read(config_name)
        try:
            self.ssl = parser.get("vars", "ssl").lower() in ["1", "true"]
        except (NoSectionError, NoOptionError):
            self.ssl = True

        if self.ssl:
            try:
                self.trust_store_path = parser.get("vars", "trust_store_path")
            except (NoSectionError, NoOptionError):
                self.trust_store_path = '/etc/pki/vdsm'
        else:
            self.trust_store_path = None
        try:
            self.management_port = parser.get("addresses", "management_port")
        except (NoSectionError, NoOptionError):
            self.management_port = '54321'

    def _get_local_vds_name(self, trusted_store_path):
        p = subprocess.Popen([
            '/usr/bin/openssl', 'x509', '-noout', '-subject', '-in',
            '%s/certs/vdsmcert.pem' % trusted_store_path], stdout=subprocess.PIPE, close_fds=True)
        out, err = p.communicate()
        if p.returncode != 0:
            return '0'
        return re.search('CN *= *([^,/$\n]+)', out.decode('UTF-8')).group(1)

    def _get_addr(self):
        return self._get_local_vds_name(self.trust_store_path)

    @staticmethod
    def _need_m2crypto():
        """Determine whether M2Crypto usage is necessary.

        If the python version supports SSL contexts and allows specifying them in xmlrpc clients, we do not need
        M2Crypto.
        """
        if 'RHSM_USE_M2CRYPTO' in os.environ and os.environ['RHSM_USE_M2CRYPTO'].lower() in ['true', '1', 'yes']:
            return False

        if not hasattr(ssl, 'create_default_context'):
            return False
        try:
            xmlrpc_client.ServerProxy('https://localhost', context=ssl.create_default_context())
            return True
        except TypeError:
            return False

    def _create_ssl_context(self):
        key_file = self.trust_store_path + '/keys/vdsmkey.pem'
        cert_file = self.trust_store_path + '/certs/vdsmcert.pem'
        ca_cert = self.trust_store_path + '/certs/cacert.pem'

        if self._need_m2crypto():
            ctx = ssl.create_default_context(capath=ca_cert)
            ctx.load_cert_chain(cert_file, key_file)
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        else:
            from M2Crypto import SSL

            ctx = SSL.Context()

            ctx.set_verify(SSL.verify_peer | SSL.verify_fail_if_no_peer_cert, 16)
            ctx.load_verify_locations(ca_cert)
            ctx.load_cert(cert_file, key_file)
        return ctx

    def connect(self):
        if self.trust_store_path:
            ssl_context = self._create_ssl_context()
            addr = self._get_local_vds_name(self.trust_store_path)
            client = self._jsonrpc(addr, ssl_context) or self._xmlrpc(addr, ssl_context)
            if client:
                return

        # Try http versions if ssl is off or fails
        self._jsonrpc('localhost') or self._xmlrpc('localhost')

    def _jsonrpc(self, addr, ssl_context=None):
        try:
            jsonrpc_client = JsonRpcClient(addr, self.management_port, ssl_context, timeout=VDSM_JSONRPC_TIMEOUT)
            jsonrpc_client.connect()
            self.jsonrpc_client = jsonrpc_client
            return True
        except IOError:
            if ssl_context:
                description = 'JSON-RPC with SSL'
            else:
                description = 'JSON-RPC'
            self.logger.warning('Unable to connect via %s' % description, exc_info=True)
        return False

    def _xmlrpc(self, addr, ssl_context=None):
        try:
            transport = None
            if ssl_context:
                if self._need_m2crypto():
                    transport = xmlrpc_client.SafeTransport(context=ssl_context)
                else:
                    from M2Crypto.m2xmlrpclib import SSL_Transport
                    transport = SSL_Transport(ssl_context)
                uri = 'https://%s:%s' % (addr, self.management_port)
            else:
                uri = 'http://%s:%s' % (addr, self.management_port)
            xmlrpc_client_lib = xmlrpc_client.ServerProxy(uri, transport)
            xmlrpc_client_lib.system.listMethods()
            self.xmlrpc_client = xmlrpc_client_lib
            return True
        except Exception:  # NOTE: unfortunately there are many ways to fail, so the exception clause is broad
            if ssl_context:
                description = 'XML-RPC with SSL'
            else:
                description = 'XML-RPC'
            self.logger.warning('Unable to connect via %s' % description, exc_info=True)
        return False

    def prepare(self):
        self.connect()

    def _get_vm_list_xmlrpc(self):
        response = self.xmlrpc_client.list(True)
        if response['status']['code'] != 0:
            self.logger.error("Unable to list virtual machines: %s", response['status']['message'])
        else:
            return response['vmList']

    def _get_vm_list_jsonrpc(self):
        try:
            return self.jsonrpc_client.call('Host.getVMList', onlyUUID=False)
        except (IOError, RuntimeError):
            self.logger.exception("Unable to list virtual machines")

    def listDomains(self):
        domains = []
        vm_list = None
        if self.jsonrpc_client is not None:
            vm_list = self._get_vm_list_jsonrpc()
        elif self.xmlrpc_client is not None:
            vm_list = self._get_vm_list_xmlrpc()
        if vm_list:
            for vm in vm_list:
                status = VDSM_STATE_TO_GUEST_STATE.get(vm['status'], Guest.STATE_UNKNOWN)
                domains.append(Guest(vm['vmId'], self.CONFIG_TYPE, status))
        return domains



from __future__ import print_function
"""
Test of VDSM virtualization backend.

Copyright (C) 2014 Radek Novacek <rnovacek@redhat.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from mock import MagicMock, patch
from unittest import TestCase

from base import TestBase

from virtwho.virt.vdsm import Vdsm
from virtwho.virt.vdsm.jsonrpc import JsonRpcClient
from virtwho.virt.vdsm.stomp import StompFrame, StompClient
from virtwho.virt.vdsm.vdsm import VdsmConfigSection
from virtwho.virt.virt import Guest

import json


class TestVdsm(TestBase):

    def create_config(self, name, wrapper, **kwargs):
        config = VdsmConfigSection(name, wrapper)
        config.update(**kwargs)
        config.validate()
        return config

    def setUp(self):
        config = self.create_config('test', None, type='vdsm')
        Vdsm._create_ssl_context = MagicMock()
        self.vdsm = Vdsm(self.logger, config, None)
        self.mock_jsonrpc_client = MagicMock()
        mock_jsonrpc_client_factory = MagicMock(return_value=self.mock_jsonrpc_client)

        self.patch_jsonrpc_client = patch('virtwho.virt.vdsm.vdsm.JsonRpcClient', mock_jsonrpc_client_factory)
        self.patch_xmlrpclib = patch('virtwho.virt.vdsm.vdsm.xmlrpc_client', MagicMock())
        self.patch_jsonrpc_client.start()
        self.patch_xmlrpclib.start()

    def tearDown(self):
        self.patch_jsonrpc_client.stop()
        self.patch_xmlrpclib.stop()

    def test_connect_via_xmlrpc(self):
        self.mock_jsonrpc_client.connect = MagicMock(side_effect=IOError)
        self.vdsm.prepare()
        self.assertIsNone(self.vdsm.jsonrpc_client)
        self.assertIsNotNone(self.vdsm.xmlrpc_client)
        self.vdsm.xmlrpc_client.list = MagicMock()
        self.vdsm.xmlrpc_client.list.return_value = {
            'status': {
                'code': 0
            },
            'vmList': [
                {
                    'vmId': '1',
                    'status': 'Down'
                }, {
                    'vmId': '2',
                    'status': 'Up'
                }, {
                    'vmId': '3',
                    'status': 'Up'
                }
            ]
        }
        domains = self.vdsm.listDomains()
        self.assertIsNone(self.vdsm.jsonrpc_client)
        self.assertEquals([d.uuid for d in domains], ['1', '2', '3'])
        self.assertEquals([d.state for d in domains], [Guest.STATE_SHUTOFF, Guest.STATE_RUNNING, Guest.STATE_RUNNING])
        self.assertEqual([d.virtWhoType for d in domains], ['vdsm', 'vdsm', 'vdsm'])
        self.vdsm.xmlrpc_client.list.assert_called_once_with(True)

    def test_connect_via_jsonrpc(self):
        self.mock_jsonrpc_client.connect = MagicMock()
        self.vdsm.prepare()
        self.assertIsNotNone(self.vdsm.jsonrpc_client)
        self.assertIsNone(self.vdsm.xmlrpc_client)
        self.vdsm.jsonrpc_client.call = MagicMock()
        self.vdsm.jsonrpc_client.call.return_value = [
            {
                'vmId': '1',
                'status': 'Down'
            }, {
                'vmId': '2',
                'status': 'Up'
            }, {
                'vmId': '3',
                'status': 'Up'
            }
        ]
        domains = self.vdsm.listDomains()
        self.assertIsNone(self.vdsm.xmlrpc_client)
        self.assertEquals([d.uuid for d in domains], ['1', '2', '3'])
        self.assertEquals([d.state for d in domains], [Guest.STATE_SHUTOFF, Guest.STATE_RUNNING, Guest.STATE_RUNNING])
        self.assertEqual([d.virtWhoType for d in domains], ['vdsm', 'vdsm', 'vdsm'])
        self.vdsm.jsonrpc_client.call.assert_called_once_with('Host.getVMList', onlyUUID=False)


class TestStomp(TestCase):
    def test_stomp_message_encoding_with_headers_without_body(self):
        connect_message = StompFrame(u'CONNECT', headers={
           u'accept-version': u'1.2',
           u'host': u'localhost',
        })
        result = connect_message.to_bytes()
        self.assertEqual(b'CONNECT\naccept-version:1.2\nhost:localhost\n\n\x00', result)

    def test_stomp_message_encoding_with_headers_with_body(self):
        connect_message = StompFrame(u'CONNECT', headers={
           u'accept-version': u'1.2',
           u'host': u'localhost',
        }, body=b'extra')
        result = connect_message.to_bytes()
        self.assertEqual(b'CONNECT\naccept-version:1.2\nhost:localhost\n\nextra\x00', result)

    def test_stomp_message_encoding_without_headers_without_body(self):
        connect_message = StompFrame(u'CONNECT')
        result = connect_message.to_bytes()
        self.assertEqual(b'CONNECT\n\n\x00', result)

    def test_stomp_message_encoding_without_headers_with_body(self):
        connect_message = StompFrame(u'CONNECT', body=b'test')
        result = connect_message.to_bytes()
        self.assertEqual(b'CONNECT\n\ntest\x00', result)

    def test_stomp_message_decoding_without_body(self):
        data = b'CONNECT\naccept-version:1.2\nhost:localhost\n\n\x00'
        connect_message = StompFrame.from_bytes(data)
        self.assertEqual(u'CONNECT', connect_message.command)
        self.assertEqual({
           u'accept-version': u'1.2',
           u'host': u'localhost',
        }, connect_message.headers)
        self.assertEqual(connect_message.body, b'')

    def test_stomp_message_decoding_with_headers_with_body(self):
        data = b'CONNECT\naccept-version:1.2\n\nlalala\x00'
        connect_message = StompFrame.from_bytes(data)
        self.assertEqual(u'CONNECT', connect_message.command)
        self.assertEqual({
            u'accept-version': u'1.2',
        }, connect_message.headers)
        self.assertEqual(connect_message.body, b'lalala')

    def test_stomp_message_decoding_without_headers_with_body(self):
        data = b'CONNECT\n\nlalala\x00'
        connect_message = StompFrame.from_bytes(data)
        self.assertEqual(u'CONNECT', connect_message.command)
        self.assertEqual({}, connect_message.headers)
        self.assertEqual(connect_message.body, b'lalala')

    def test_stomp_message_decoding_without_headers_without_body(self):
        data = b'CONNECT\n\n\x00'
        connect_message = StompFrame.from_bytes(data)
        self.assertEqual(u'CONNECT', connect_message.command)
        self.assertEqual({}, connect_message.headers)
        self.assertEqual(connect_message.body, b'')


class StompClientTest(TestCase):
    def test_stomp_client_connect_sends_connect_frame(self):
        with patch('socket.socket') as socket_factory:
            mock_socket = MagicMock()
            socket_factory.return_value = mock_socket
            mock_socket.recv.return_value = b'CONNECTED\n\n\x00'
            client = StompClient('localhost', 54321)
            client.connect()
        mock_socket.sendall.assert_called_once_with(b'CONNECT\naccept-version:1.2\nhost:localhost\n\n\x00')

    def test_stomp_client_cleans_up_subscriptions_on_close(self):
        with patch('socket.socket') as socket_factory:
            with patch('uuid.uuid4') as mock_uuid:
                mock_uuid.return_value = '42'
                mock_socket = MagicMock()
                socket_factory.return_value = mock_socket
                mock_socket.recv.return_value = b'CONNECTED\n\n\x00'
                client = StompClient('localhost', 54321)
                client.connect()
                client.subscribe('bob')
                client.close()
        mock_socket.sendall.assert_called_with(b'UNSUBSCRIBE\nid:42\n\n\x00')


class JsonRpcTest(TestCase):
    def test_jsonrpc_client_subscribes_to_response_queue(self):
        with patch('virtwho.virt.vdsm.jsonrpc.StompClient') as mock_client_factory:
            mock_stomp_client = MagicMock()
            mock_client_factory.return_value = mock_stomp_client
            jsonrpc_client = JsonRpcClient('localhost', '54321')
            jsonrpc_client.connect()
            mock_stomp_client.subscribe.assert_called_once_with(jsonrpc_client.id)

    def test_jsonrpc_client_payload_encoding_with_params(self):
        with patch('virtwho.virt.vdsm.jsonrpc.StompClient') as mock_client_factory:
            with patch('uuid.uuid4') as mock_uuid:
                mock_uuid.return_value = 42
                mock_stomp_client = MagicMock()
                mock_client_factory.return_value = mock_stomp_client
                mock_stomp_client.receive.return_value = StompFrame('MESSAGE', body=b'{"result":"bar"}')
                jsonrpc_client = JsonRpcClient('localhost', '54321')
                jsonrpc_client.connect()
                result = jsonrpc_client.call('test', foo='bar')
                self.assertEquals(result, u'bar')
                expected_command = u'SEND'
                expected_headers = {
                    u'content-length': 74,
                    u'destination': u'jms.topic.vdsm_requests',
                    u'reply-to': u'42'
                }
                expected_data = {
                    "params": {"foo": "bar"}, "jsonrpc": "2.0", "method": "test", "id": "42"
                }

                # The data in the stomp_client.send call (2nd, and 3rd arguments) are dicts dumped
                # as json to a bytes object. Because dictionaries are not guaranteed to be iterated
                # over in the same order, these dicts do not appear equal on python 3.
                # Comparing these using assertEqual will call the registered method associated with
                # dicts (which does not care about order).
                send_args = mock_stomp_client.send.call_args
                self.assertEqual(send_args[0][0], expected_command)
                self.assertEqual(send_args[0][1], expected_headers)
                actual_data = json.loads(send_args[0][2])
                self.assertEqual(actual_data, expected_data)

    def test_jsonrpc_client_payload_encoding_without_params(self):
        with patch('virtwho.virt.vdsm.jsonrpc.StompClient') as mock_client_factory:
            with patch('uuid.uuid4') as mock_uuid:
                mock_uuid.return_value = 42
                mock_stomp_client = MagicMock()
                mock_client_factory.return_value = mock_stomp_client
                mock_stomp_client.receive.return_value = StompFrame('MESSAGE', body=b'{"result":"bar"}')
                jsonrpc_client = JsonRpcClient('localhost', '54321')
                jsonrpc_client.connect()
                result = jsonrpc_client.call('test')
                self.assertEqual(result, u'bar')
                mock_stomp_client.send.assert_called_with(u'SEND', {
                    u'content-length': 48,
                    u'destination': u'jms.topic.vdsm_requests',
                    u'reply-to': u'42'
                }, b'{"jsonrpc": "2.0", "method": "test", "id": "42"}')

    def test_jsonrpc_client_errs_on_error_frame(self):
        with patch('virtwho.virt.vdsm.jsonrpc.StompClient') as mock_client_factory:
            with patch('uuid.uuid4') as mock_uuid:
                mock_uuid.return_value = 42
                mock_stomp_client = MagicMock()
                mock_client_factory.return_value = mock_stomp_client
                mock_stomp_client.receive.return_value = StompFrame('ERROR', body=b'uh-oh!')
                jsonrpc_client = JsonRpcClient('localhost', '54321')
                jsonrpc_client.connect()
                self.assertRaises(RuntimeError, jsonrpc_client.call, 'test')

    def test_jsonrpc_client_errs_on_error_json(self):
        with patch('virtwho.virt.vdsm.jsonrpc.StompClient') as mock_client_factory:
            with patch('uuid.uuid4') as mock_uuid:
                mock_uuid.return_value = 42
                mock_stomp_client = MagicMock()
                mock_client_factory.return_value = mock_stomp_client
                mock_stomp_client.receive.return_value = StompFrame('MESSAGE', body=b'{"error":{"message":"foo"}}')
                jsonrpc_client = JsonRpcClient('localhost', '54321')
                jsonrpc_client.connect()
                self.assertRaises(RuntimeError, jsonrpc_client.call, 'test')
